import os
import glob
import re
from .tempfiles import create_temp_file
from ..utils.types import Camera
from typing import List

from ..utils import import_module

ffmpeg = import_module("ffmpeg", "ffmpeg-python")


valid_frames_exts = ("png", "jpg", "jpeg", "bmp", "tiff", "tif")


def get_frames_from_directory(directory, exts=valid_frames_exts, sort=True):
    """Return a list of all paths to files in directory which have a valid frame extension"""
    files = [
        os.path.join(directory, f)
        for f in os.listdir(directory)
        if os.path.splitext(f)[-1][1:] in exts
    ]
    if sort:
        files = sorted(files)
    return files


def ffmpeg_input_from_files(file_list, frame_rate=25):
    """Return ffmpeg input from list of files"""
    file_list_loc = create_temp_file(ext=".txt")

    with open(file_list_loc, "w") as f:
        for file in file_list:
            abs_file = os.path.abspath(file)
            f.write(f"file '{abs_file}'\n")

    return ffmpeg.input(file_list_loc, r=str(frame_rate), f="concat", safe="0")


def frames_to_video(
    *args,
    frame_list=None,
    frame_fmt=None,
    directory=None,
    frame_rate=30,
    output_loc="./output.mp4",
    output_fmt="mp4",
    pix_fmt="yuv420p",
    overwrite=True,
    loglevel="error",
    delete_images=False,
):
    """
    Need to input one of frame_list or frame_fmt or directory

    :param frame_list: [optional] list of frames to concatenate into a video
    :param frame_fmt: [optional] frame format to use as input
    :param directory: [optional] directory to use as input (loads sorted images in directory)

    :param frame_rate: frame rate of video
    :param output_loc: video file location
    :param output_fmt: video file format
    :param pix_fmt: pixel format (see ffmpeg documentation, default yuv420p)

    :param overwrite: overwrite existing video file
    :param loglevel: ffmpeg loglevel (see ffmpeg documentation, default error)

    :param delete_images: delete images after video is created
    :return:
    """

    assert args == (), "frames_to_video() takes keyword arguments only"

    # Assert exactly 1 of frame_list, frame_fmt, directory is provided
    assert (
        sum([frame_list is not None, frame_fmt is not None, directory is not None]) == 1
    ), "Exactly one of frame_list, frame_fmt, directory must be provided"

    vid_input = None
    file_list = []
    if frame_list:
        assert frame_list, "No frames provided"
        vid_input = ffmpeg_input_from_files(frame_list, frame_rate=frame_rate)
        file_list = frame_list

    elif frame_fmt:
        vid_input = ffmpeg.input(frame_fmt)

        # Get file list using glob (note: this does not match the exact frame format, so may not be fully robust)
        file_list = glob.glob(re.sub(r"%\d*[df]", "*", frame_fmt))

    elif directory:
        file_list = get_frames_from_directory(directory)
        assert file_list, "No valid frames found in directory"

        vid_input = ffmpeg_input_from_files(file_list, frame_rate=frame_rate)

    (
        vid_input.output(output_loc, format=output_fmt, r=frame_rate, pix_fmt=pix_fmt)
        .global_args("-loglevel", loglevel)
        .run(overwrite_output=overwrite)
    )

    if delete_images:
        for f in file_list:
            os.remove(f)

    print("Video created at " + output_loc)


def frames_to_video_multiview(
    cameras: List[Camera], directory=".", out_loc="vid.mp4", **kwargs
):
    """
    Following the rendering of a multiview animation, this function handles rendering the frames separately into
    videos.

    :param cameras: List of cameras used for rendering
    :param directory: Directory to look for frames in
    :param out_loc: Output location for videos. Each camera will be saved as a separate video with the camera name prepended
    :param kwargs: Keyword arguments to pass to :func:`frames_to_video`
    """

    for cam in cameras:
        frame_files = sorted(
            [
                os.path.join(directory, f)
                for f in os.listdir(directory)
                if f.startswith(cam.name)
            ]
        )
        out_f = os.path.join(
            os.path.dirname(out_loc), f"{cam.name}_{os.path.basename(out_loc)}"
        )
        frames_to_video(frame_list=frame_files, output_loc=out_f, **kwargs)



"""For calling errors if a user tries run certain operations in the wrong python environment"""

from importlib import import_module
from typing import Union

package_name = "blendersynth"


def conditional_import(
    condition: bool, src: str, name: str = None
) -> Union["IllegalImport", object]:
    """
    [No longer used in __init__.py, but kept here for reference]
    Imports a module if condition,
    otherwise returns IllegalImport.

    :param condition: If True, import the module
    :param src: The module to import
    :param name: The name of the module to import. If None, the last part of src is used.
    :return: The imported module, or IllegalImport if condition is False
    """

    if condition:
        if src.startswith("."):
            module = import_module(src, package=package_name)
        else:
            module = import_module(src)

        if name is None:
            return module

        return getattr(module, name if name else src.split(".")[-1])
    else:
        return IllegalImport()


class IllegalImport:
    """Object which raises error on any interaction, to tell users they have used BlenderSynth incorrectly."""

    def __init__(self, REQUIRES_BLENDER=True):
        self.REQUIRES_BLENDER = REQUIRES_BLENDER

    def __call__(self, *args, **kwargs):
        raise ImportError(self._message)

    def __getattr__(self, name):
        raise ImportError(self._message)

    @property
    def _message(self):
        if self.REQUIRES_BLENDER:
            return (
                "This object can only be used inside a Blender python environment. "
                "Make sure that either:"
                "\n- bsyn.run_this_script() is run on a previous line"
                "\n- This script is being run from another python script, using bsyn.execute_jobs."
            )

        else:
            return "This object can only be used outside a Blender python environment. "



"""Functions for boolean operations of meshes"""
import bpy
from ..utils import types


def _apply_boolean(
    target: types.Mesh,
    *others: types.Mesh,
    operation: str = "UNION",
    hide_others: bool = True
) -> types.Mesh:
    for other in others:
        target.add_child(other)

        modifier = target.obj.modifiers.new(type="BOOLEAN", name="bool")
        modifier.object = other.obj
        modifier.operation = operation

        bpy.ops.object.modifier_apply(modifier="bool")
        if hide_others:
            other.obj.hide_render = True
            other.obj.hide_set(True)
    return target


def union(target: types.Mesh, *others: types.Mesh, hide_others=True) -> types.Mesh:
    """Apply the union boolean modifier to a collection of meshes.

    :param target: The target Mesh: this is returned by the function
    :param others: Other Meshes to union with `target`
    :param hide_others: If True, hide all Meshes in the `others` list"""
    return _apply_boolean(target, *others, operation="UNION", hide_others=hide_others)


def difference(target: types.Mesh, *others: types.Mesh, hide_others=True) -> types.Mesh:
    """Apply the difference boolean modifier to a collection of meshes.

    :param target: The target Mesh: this is returned by the function
    :param others: Other Meshes to subtract from `target`
    :param hide_others: If True, hide all Meshes in the `others` list"""

    return _apply_boolean(
        target, *others, operation="DIFFERENCE", hide_others=hide_others
    )


def intersect(target: types.Mesh, *others: types.Mesh, hide_others=True) -> types.Mesh:
    """Apply the intersect boolean modifier to a collection of meshes.

    :param target: The target Mesh: this is returned by the function
    :param others: Other Meshes to intersect with `target`
    :param hide_others: If True, hide all Meshes in the `others` list"""
    return _apply_boolean(
        target, *others, operation="INTERSECT", hide_others=hide_others
    )



import bpy
from .bsyn_object import BsynObject
from .nodes import tidy_tree
from ..utils import io
from ..utils import version


def _new_shader_node(node_tree, node_type):
    shader_node = node_tree.nodes.new(type=node_type)
    if node_type == "ShaderNodeBsdfPrincipled":
        spec_key = "Specular IOR Level" if version.is_version_plus(4) else "Specular"
        shader_node.inputs[spec_key].default_value = 0

    return shader_node


class Material(BsynObject):
    """BlenderSynth Material class. Will always be a node material."""

    def __init__(
        self, name="NewMaterial", shader_type="ShaderNodeBsdfPrincipled", mat=None
    ):
        if mat is None:
            mat = bpy.data.materials.new(name)
            mat.use_nodes = True
            mat.node_tree.nodes.clear()
            self.shader = _new_shader_node(mat.node_tree, shader_type)
            self.output = mat.node_tree.nodes.new(type="ShaderNodeOutputMaterial")
            mat.node_tree.links.new(
                self.shader.outputs[0], self.output.inputs["Surface"]
            )

        else:
            self.shader = None
            self.output = mat.node_tree.nodes["Material Output"]

        mat.use_nodes = True
        self._object = mat

        self.node_tree = mat.node_tree
        self.nodes = self.node_tree.nodes
        self.links = self.node_tree.links

        self._image_nodes = {}  # dict of shader input socket : image node

        # set-up scaling
        self._add_scaling_node()
        self._scale = 1

        tidy_tree(self.node_tree)

    @classmethod
    def from_image(
        cls,
        image_loc,
        name="ImageMaterials",
        shader_type="ShaderNodeBsdfPrincipled",
        use_global_scale=True,
    ) -> "Material":
        """
        Create a new material from an image texture.
        :param image_loc: Location of the image
        :param name: Name of the material
        :param shader_type: Shader type
        :param use_global_scale: Flag to scale the image texture node by the global scale
        :return: New Material instance
        """
        mat = cls(name, shader_type=shader_type)  # new instance
        mat.add_source(
            image_loc, input_name="Base Color", use_global_scale=use_global_scale
        )

        return mat

    def add_source(
        self,
        image_loc: str,
        input_name: str = "Base Color",
        use_global_scale: bool = True,
        allow_duplicate: bool = False,
    ):
        """
        Add an image texture node to the material node tree, and connect it to the specified input of the Shader.
        :param image_loc: Image location
        :param input_name: Input socket name of the Shader to connect the image texture node to
        :param use_global_scale: Flag to scale the image texture node by the global scale
        :param allow_duplicate: Flag to allow duplicate image texture nodes (by default, a node to the same input will overwrite any previous)
        :return:
        """
        # Get the nodes in the material node tree
        nodes = self.obj.node_tree.nodes
        links = self.obj.node_tree.links

        # Find the Principled BSDF
        principled = next(
            (node for node in nodes if node.type == "BSDF_PRINCIPLED"), None
        )
        if not principled:
            raise ValueError(
                "add_source requires a Principled BSDF node in the material node tree"
            )

        if input_name in self._image_nodes and not allow_duplicate:
            tex_image_node = self._image_nodes[input_name]
        else:
            tex_image_node = nodes.new("ShaderNodeTexImage")

        tex_image_node.image = io.load_image(image_loc)
        self._image_nodes[input_name] = tex_image_node

        if input_name == "Displacement":
            # Connect to the 'Displacement' socket of the output node
            links.new(
                tex_image_node.outputs["Color"], self.output.inputs["Displacement"]
            )

        else:
            # Connect the image texture node to the specified input of the Principled BSDF
            links.new(tex_image_node.outputs["Color"], principled.inputs[input_name])

        if use_global_scale:
            links.new(self._texture_scaling_socket, tex_image_node.inputs["Vector"])

        if "color" not in input_name.lower():
            tex_image_node.image.colorspace_settings.name = "Non-Color"

        tidy_tree(self.node_tree)

    @classmethod
    def from_blender_material(cls, blender_mat: bpy.types.Material):
        if not isinstance(blender_mat, bpy.types.Material):
            raise ValueError("Invalid Blender material")

        # Create a new Material instance
        mat = cls(blender_mat.name, mat=blender_mat)

        # Find the shader and output nodes in the copied node tree
        mat.shader = next(
            (node for node in mat.nodes if node.type == "BSDF_PRINCIPLED"), None
        )
        mat.output = next(
            (node for node in mat.nodes if node.type == "OUTPUT_MATERIAL"), None
        )

        return mat

    def _add_scaling_node(self):
        """Add scaling node to the material node tree"""

        self.scale_node = self.nodes.new("ShaderNodeValue")
        tex_coord_node = self.nodes.new("ShaderNodeTexCoord")

        mapping_node = self.nodes.new("ShaderNodeMapping")

        self.links.new(self.scale_node.outputs[0], mapping_node.inputs["Scale"])
        self.links.new(tex_coord_node.outputs["UV"], mapping_node.inputs["Vector"])
        self._texture_scaling_socket = mapping_node.outputs["Vector"]
        self.scale = 1  # set default scale

    def set_bdsf_property(self, key: str, value: float):
        """Set the property of the BSDF node"""
        if self.shader.type != "BSDF_PRINCIPLED":
            raise ValueError("Shader is not a Principled BSDF")

        if key not in self.shader.inputs:
            raise ValueError(f"Invalid BDSF shader property `{key}`")

        self.shader.inputs[key].default_value = value

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, value):
        self._scale = value
        self.scale_node.outputs[0].default_value = value

    @property
    def name(self):
        return self.obj.name



from typing import Union, List, Tuple, ForwardRef

try:
    from mathutils import Vector
except ImportError:
    Vector = "mathutils.Vector"

import numpy as np

types_loc = "blendersynth.utils.types"

VectorLike = Union[Vector, np.ndarray, List, Tuple]
"""A type hint that represents vector-like objects. Can be any of:

- :class:`mathutils.Vector`
- :class:`numpy.ndarray`
- :class:`list`
- :class:`tuple`
"""

VectorLikeOrScalar = Union[VectorLike, int, float]
"""A type hint that represents vector-like objects. Can be any of:

- :class:`~VectorLike`
- :class:`int`
- :class:`float`
"""

## Annotations
KeypointOrAxesAnnotation = List[np.ndarray]
"""Keypoint Annotations, an N x 2 ndarray per instance,
or Axes annotation, a 4 x 2 ndarray per instance"""

BboxAnnotation = List[VectorLike]
"""Bbox annotation, a 1D array per instance"""


# convert our type hint to Sphinx so it can be searched for in documentaton
def sphinxify_type_hint(py_type):
    if py_type is List:  # Handle built-in types
        return ":py:class:`typing.List`"
    if py_type is Tuple:
        return ":py:class:`typing.Tuple`"
    if py_type is int:
        return ":py:class:`int`"
    if py_type is float:
        return ":py:class:`float`"

    if isinstance(py_type, ForwardRef):
        return f":py:class:`~{py_type.__forward_arg__}`"

    elif isinstance(py_type, str):
        return f":py:class:`~{py_type}`"

    elif hasattr(py_type, "__origin__") and hasattr(py_type, "__args__"):
        origin = py_type.__origin__
        args = py_type.__args__
        sphinx_args = ", ".join(sphinxify_type_hint(arg) for arg in args)

        if origin is Union:
            origin_name = "Union"
        else:
            origin_name = (
                origin.__name__
                if hasattr(origin, "__name__")
                else str(origin).split(".")[-1]
            )

        return f":py:data:`~typing.{origin_name}`[{sphinx_args}]"

    else:
        return f":py:class:`~{py_type.__module__}.{py_type.__name__}`"


sphinx_mappings = (
    {}
)  # dictionary of Sphinx auto type hint -> Type hint within this file
wrapper_mappings = {}  # dictionary of object -> type hint within this file
for k in [
    "VectorLike",
    "VectorLikeOrScalar",
    "KeypointOrAxesAnnotation",
    "BboxAnnotation",
]:
    obj = globals()[k]
    key = sphinxify_type_hint(obj)
    value = f":class:`{k} <{types_loc}.{k}>`"
    sphinx_mappings[key] = value
    wrapper_mappings[obj] = value

# to avoid circular referencing, we also create some more type hints here (e.g. Mesh)
Mesh = "blendersynth.blender.mesh.Mesh"
Camera = "blendersynth.blender.camera.Camera"

for name in ["Mesh"]:
    obj = globals()[name]
    sphinx_mappings[sphinxify_type_hint(obj)] = f":class:`{name} <{obj}>`"



import tempfile
import os

tempdir = os.path.join(tempfile._get_default_tempdir(), "blendersynth")
os.makedirs(tempdir, exist_ok=True)


def create_temp_file(ext=".png"):
    """Create and return a temporary filename that will be in the user's tempdir.
    These can all be cleaned up after use by the `cleanup_temp_files` function.

    :param ext: file extension to use"""

    if not ext.startswith("."):
        ext = "." + ext

    return tempfile.mktemp(dir=tempdir) + ext


def cleanup_temp_files():
    """Delete all temporary files in the `blendersynth` tempdir"""
    for f in os.listdir(tempdir):
        os.remove(os.path.join(tempdir, f))



import subprocess
import os
import sys
from blendersynth.utils.blender_setup.blender_locator import (
    find_blender_python,
    get_blender_path,
    remove_config,
    write_to_config,
    read_from_config,
    remove_from_config,
    set_blender_path,
)

dependencies = [
    "imageio",
    "numpy",
    "platformdirs",
    "tqdm",
    "opencv-python",
    "ffmpeg-python",
    "pyyaml",
    "pydevd-pycharm",
    "debugpy",  # for debugging
]


def setup_blender_stubs(blender_path):
    """Install blender stubs in vanilla python environment,
    for correct Blender version"""

    vanilla_python_executable = sys.executable

    # Get Blender version from the output
    script_code = "import bpy; print('VERSION' + '.'.join(map(str, bpy.app.version)))"
    result = subprocess.run(
        [blender_path, "--background", "--python-expr", script_code],
        capture_output=True,
        text=True,
    )

    for line in result.stdout.split("\n"):
        if line.startswith("VERSION"):
            blender_version = line[len("VERSION") :].strip()
            break
    else:
        raise Exception("Could not get Blender version")

    # Install stubs
    # Should really use blender_version here
    if not check_module(vanilla_python_executable, "blender-stubs"):
        _install_module(vanilla_python_executable, "blender-stubs")


def check_module(python_executable, module_name):
    try:
        subprocess.check_call(
            [python_executable, "-m", "pip", "show", module_name],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
        )
        return True
    except subprocess.CalledProcessError:
        return False


def _install_module(
    python_executable,
    module_name,
    is_test_pypi=False,
    version=None,
    upgrade=True,
    editable=False,
):
    commands = [python_executable, "-m", "pip"]
    commands += ["install"]
    if is_test_pypi:
        commands += ["-i", "https://test.pypi.org/simple/"]

    if editable:
        commands += ["-e"]

    commands += [module_name + (f"=={version}" if version is not None else "")]
    if upgrade:
        commands += ["--upgrade"]

    try:
        subprocess.check_call(commands)
    except subprocess.CalledProcessError as e:
        raise Exception(f"Could not install {module_name} via pip. Error: {e}")


def install_module(
    module_name: str, version: str = None, upgrade: bool = True, editable: bool = False
):
    """Install a module to Blender Python.

    :param module_name: name of module to install
    :param version: version of module to install
    :param upgrade: if True, will upgrade module if it is already installed
    :param editable: if True, will install module in editable mode
    """

    python_exec = find_blender_python(get_blender_path())

    _install_module(
        python_exec, module_name, version=version, upgrade=upgrade, editable=editable
    )


def check_blender_install(
    force_all=False,
    force_find_blender=False,
    force_find_blender_python=False,
    force_install_dependencies=False,
    blendersynth_from_local=False,
    blendersynth_editable=False,
):
    """Check if Blender is installed correctly and has all necessary packages.
    If not, run first time setup.

    On first time setup, will create a file, config.ini, in the user's config,
    containing the necessary info.

    Force: if True, will run first time setup (overwriting any existing config.ini)
    regardless"""

    if force_all:
        remove_config()  # remove config file if it exists to force first time setup

    if force_find_blender:
        remove_from_config("BLENDER_PATH")
    if force_find_blender_python:
        remove_from_config("BLENDER_PYTHON_PATH")
    if force_install_dependencies:
        remove_from_config("DEPENDENCIES_INSTALLED")

    blender_path = get_blender_path()
    if blender_path is None:
        blender_path = set_blender_path()

    setup_blender_stubs(blender_path)  # check blender stubs

    python_path = find_blender_python(blender_path)

    if not read_from_config("DEPENDENCIES_INSTALLED") == "True":
        # check if blender's python has all necessary packages
        for dependency in dependencies:
            if not check_module(python_path, dependency):
                _install_module(python_path, dependency)

        # First uninstall blendersynth if present (in case of updates)
        if check_module(python_path, "blendersynth"):
            subprocess.check_call(
                [python_path, "-m", "pip", "uninstall", "-y", "blendersynth"]
            )

        # Install blendersynth package to blender's python
        if blendersynth_from_local:
            # Install from local setup.py
            setup_py_loc = os.path.join(
                os.path.dirname(__file__), "..", "..", "..", "setup.py"
            )
            directory_loc = os.path.dirname(setup_py_loc)
            _install_module(python_path, directory_loc, editable=blendersynth_editable)

        else:
            # Install from pypi
            _install_module(python_path, "blendersynth", upgrade=True)

        write_to_config("DEPENDENCIES_INSTALLED", "True")



"""Some preferential blender operations to run (e.g. setup) that will be run before debug mode is entered"""
import bpy
from ..blender import render


def on_script_open(delete_existing=True):
    """
    On script open:

    - Set Engine to CYCLES
    - Delete default Cube & Light
    - Disable splash screen
    """
    render.set_engine("CYCLES")

    if delete_existing:
        bpy.data.objects["Cube"].select_set(True)
        bpy.data.objects["Light"].select_set(True)
        bpy.ops.object.delete()

    bpy.context.preferences.view.show_splash = False  # disable splash screen



import sys
from tqdm import tqdm
from time import perf_counter, sleep
import numpy as np
from subprocess import Popen
import atexit

from datetime import datetime, timedelta
import os


def _list_split(list, chunks):
    """Split a list into chunks"""
    return [[*x] for x in np.array_split(list, chunks)]


class BlenderThread:
    """Manages a list of jobs, which it feeds into sequential Blender instances in a single thread."""

    def __init__(
        self,
        command,
        jobs,
        log_loc,
        progress_loc,
        name="",
        timeout: int = 100,
        to_stdout: bool = False,
        MAX_PER_JOB: int = 100,
        script_directory: str = None,
    ):
        """
        :param progress_loc: .log file to write progress to
        :param timeout: longest time in (s) without render after which process is finished/failed.
        :param to_stdout: If True, print to stdout instead of to a log file.
        :param MAX_PER_JOB: Split command into jobs of size MAX_PER_JOB, and run each job in a separate process.
        :param script_directory: If given, add this to `sys.path` before running the script.
        """

        self.command = command
        self.jobs = _list_split(
            jobs, np.ceil(len(jobs) / MAX_PER_JOB)
        )  # split jobs into chunks of MAX_PER_JOB
        self.njobs = len(self.jobs)

        self.size = len(jobs)
        self.timeout = timeout
        self.script_directory = script_directory

        self.prev_n = (
            0  # store how many rendered to check for updates for timeout purposes
        )
        self.timer = perf_counter()

        self.to_stdout = to_stdout
        if not self.to_stdout and log_loc is not None:
            self.log_loc = log_loc
            self.logfile = open(self.log_loc, "a")
        else:
            self.log_loc, self.logfile = None, None

        self.logger_loc = progress_loc

        self.process = None
        self.job = -1
        self.finished = False

        self.name = name
        self.status = f"STARTED THREAD {self.name}"

    def check_in(self):
        """Checks if current job still running, if not move to next job.
        If all jobs complete, set self.finished = True"""
        if self.is_running:
            return

        self.job += 1

        if self.job >= self.njobs:
            self.finished = True
            self.status = f"✓ THREAD {self.name} COMPLETED."
            return

        self.status = f"THREAD {self.name} RUNNING JOB {self.job + 1}/{self.njobs}..."
        self.start_job(self.job)

    def start_job(self, job: int = 0):
        """
        :param job: Index of job to start
        """
        job_list = self.jobs[job]
        command = self.command.set_job(job_list).set_logger(self.logger_loc).command

        stdout = sys.stdout if self.to_stdout else self.logfile
        stderr = sys.stderr if self.to_stdout else self.logfile

        env = os.environ.copy()
        if self.script_directory is not None:
            env["PYTHONPATH"] = (
                self.script_directory + os.pathsep + env.get("PYTHONPATH", "")
            )

        self.process = Popen(
            command, universal_newlines=True, stdout=stdout, stderr=stderr, env=env
        )

    def terminate(self):
        """End process"""
        self.process.kill()

    def kill(self):
        self.terminate()

    @property
    def is_running(self):
        if self.process:
            return self.process.poll() is None
        return False

    @property
    def complete(self):
        if not self.is_running:
            self.logfile.flush()
            return True

    def __len__(self):
        return self.size

    @property
    def num_rendered(self):
        """Read through Log file to find how many renders have been completed"""
        if self.process is None or not os.path.isfile(self.logger_loc):
            return 0  # process not started yet

        x = 0
        with open(self.logger_loc, "r") as f:
            for line in f.readlines():
                x += 1  # currently only 1 type of message to logger - render complete

        return x

    def remaining_idxs(self):
        raise NotImplementedError()

    @property
    def success(self):
        return self.num_rendered == self.size

    def check_status(self):
        """True if still running successfully, False if exceeded timeout"""
        n = self.num_rendered
        if n > self.prev_n:
            self.prev_n = n
            self.timer = perf_counter()

        elif (perf_counter() - self.timer) >= self.timeout:
            self.status = f"✖ THREAD {self.name} FAILED [TIMEOUT]."
            return False

        return True


class BlenderThreadManager:
    """Manager of multiple :class:`~blendersynth.run.blender_threading.BlenderThread` instances"""

    def __init__(
        self,
        command,
        jsons,
        output_directory,
        print_to_stdout=False,
        MAX_PER_JOB=100,
        script_directory=None,
    ):
        """
        :param commands: Base Blender command to run
        :param jsons: A list of num_threads size, each element is a list of jsons to render from
        :param log_locs:
        :param MAX_PER_JOB: To prevent memory issues, split up jobs into chunks of MAX_PER_JOB
        :param script_directory: If given, add this to `sys.path` before running the script.
        """
        self.num_threads = len(jsons)

        self.command = command

        # create logs
        session_name = datetime.now().strftime(r"%y%m%d-%H%M%S")
        log_dir = os.path.join(output_directory, "logs", session_name)
        progress_dir = os.path.join(log_dir, "_progress")  # for storing progress

        os.makedirs(log_dir, exist_ok=True)
        os.makedirs(progress_dir, exist_ok=True)

        logs = [
            os.path.join(log_dir, f"log_{i:02d}.txt") for i in range(self.num_threads)
        ]
        progresses = [
            os.path.join(progress_dir, f"progress_{i:02d}.log")
            for i in range(self.num_threads)
        ]
        self.log_locs = logs

        # Set report name as report_xx, incrementing by 1 each report
        report_fname = f"report_{len([f for f in os.listdir(output_directory) if 'report' in f]):02d}.txt"
        self.report_loc = os.path.join(output_directory, report_fname)

        self.t0 = 0
        self.session_start = None

        self.threads = []
        for i in range(self.num_threads):
            thread = BlenderThread(
                command,
                jobs=jsons[i],
                log_loc=None if logs is None else logs[i],
                progress_loc=progresses[i],
                name=str(i),
                to_stdout=print_to_stdout,
                MAX_PER_JOB=MAX_PER_JOB,
                script_directory=script_directory,
            )

            self.threads.append(thread)

    def __len__(self):
        return sum(map(len, self.threads))

    def start(
        self,
        progress_bars: bool = True,
        tick: float = 0.5,
        report_every: float = 15.0,
        offset: float = 1.0,
    ):
        """Start all threads and job progress
        :param progress_bars: Show progress bars for each thread and overall progress
        :param tick: How often to update progress bars
        :param report_every: How often to print status updates to log
        :param offset: How long to wait before starting each thread (to avoid memory issues)
        """
        self.t0 = perf_counter()  # log start time
        self.session_start = datetime.now()

        for thread in self.threads:
            thread.check_in()
            sleep(offset)

        # register atexit handler
        atexit.register(self.terminate)

        last_report_time = perf_counter()

        if progress_bars:
            pbars = []
            bar_format = "{l_bar}{bar:10}{r_bar}{bar:-10b}"
            for i, thread in enumerate(self.threads):
                p = tqdm(
                    total=len(thread),
                    initial=thread.num_rendered,
                    bar_format=bar_format,
                    position=i + 1,
                )
                pbars.append(p)

            with tqdm(total=len(self), bar_format=bar_format, position=0) as pbar:
                while any(t.is_running for t in self.threads):
                    sleep(tick)

                    # Update progress bar
                    rendered_images = self.num_rendered

                    desc = "Session rendering..."
                    desc += f" [Dataset: {rendered_images}/{len(self)}]"
                    pbar.set_description(desc)

                    pbar.n = self.num_rendered
                    pbar.refresh()

                    if (perf_counter() - last_report_time) >= report_every:
                        self.update_report()
                        last_report_time = perf_counter()

                    # Update sub-progress bars, restart threads if needed
                    for t, thread in enumerate(self.threads):
                        if thread.finished:
                            continue

                        thread.check_in()

                        if thread.success:
                            pbars[t].n = thread.num_rendered

                        else:
                            thread_running = thread.check_status()

                            if thread_running:
                                pbars[t].n = thread.num_rendered

                            else:  # Thread failed, currently will only notify the user
                                pass

                        pbars[t].set_description(thread.status)

            self.update_report()

    @property
    def num_rendered(self):
        return sum(t.num_rendered for t in self.threads)

    def update_report(self):
        t1 = perf_counter()
        elapsed = t1 - self.t0

        report = []
        if self.num_rendered > 0:
            # calculate number of seconds remaining
            s_remaining = (len(self) - self.num_rendered) * (
                elapsed / self.num_rendered
            )
            eta = datetime.now() + timedelta(seconds=s_remaining)

            report += [
                f"Number of images rendered: {self.num_rendered}\n",
                f"Total session quota: {len(self)}\n",
                f"Time elapsed: {timedelta(seconds=round(elapsed))}\n",
                f"Time per render (s): {elapsed / self.num_rendered:.2f}\n\n",
                f"Session start: {self.session_start.strftime('%I:%M %p %d/%m/%y')}\n"
                f"Estimated End: {eta.strftime('%I:%M %p %d/%m/%y')}",
            ]

        with open(self.report_loc, "w") as outfile:
            outfile.writelines(report)

    def terminate(self):
        """End all threads"""
        for thread in self.threads:
            thread.terminate()



import bpy
import numpy as np
import mathutils
from .utils import handle_vec, SelectObjects, animatable_property
from .bsyn_object import BsynObject
from typing import Union
from .mesh import Mesh
from .curve import Curve
from ..utils import types


def _look_at_rotation(
    obj_camera: bpy.types.Object,
    at: mathutils.Vector = mathutils.Vector((0, 0, 0)),
    up: mathutils.Vector = mathutils.Vector((0, 0, 1)),
) -> mathutils.Euler:
    """Rotate camera to look at 'at', with 'up' maintained.

    :param obj_camera: Camera object
    :param at: Point to look at
    :param up: Up vector"""

    camera_position = obj_camera.location

    z_axis = -(
        at - camera_position
    ).normalized()  # Blender forwards is '-z', so use -ve here
    x_axis = up.cross(z_axis)
    y_axis = z_axis.cross(x_axis).normalized()
    is_close = all(abs(i) < 5e-3 for i in x_axis)

    if is_close:
        replacement = y_axis.cross(z_axis).normalized()
        x_axis = replacement

    R = mathutils.Matrix([x_axis, y_axis, z_axis]).transposed()
    euler = R.to_euler()
    return euler


class Camera(BsynObject):
    """Camera object, to handle movement, tracking, etc."""

    def __init__(self, camera: bpy.types.Object = None):
        if camera is None:
            camera = bpy.context.scene.camera

        self.camera = camera
        self._object = camera

    @classmethod
    def create(
        cls,
        name: str = "Camera",
        location: Union[list, tuple, np.ndarray] = None,
        rotation: Union[list, tuple, np.ndarray] = None,
    ) -> "Camera":
        """Create a new camera object.

        :param name: Name of camera
        :param location: Location of camera
        :param rotation_euler: Rotation of camera in euler angles"""

        with SelectObjects():  # revert selection after creating camera
            bpy.ops.object.camera_add(enter_editmode=False, align="VIEW")
            camera = bpy.context.object
            camera.name = name

        cam = cls(camera)

        if location is not None:
            cam.set_location(location)

        if rotation is not None:
            cam.set_rotation_euler(rotation)

        return cam

    @classmethod
    def from_scene(cls, name: str = "Camera") -> "Camera":
        """Create from a named camera in scene

        :param name: Name of camera"""
        assert name in bpy.data.objects, f"Camera {name} does not exist"
        return cls(bpy.data.objects[name])

    def update(self):
        """Update view layer and depsgraph"""
        bpy.context.view_layer.update()
        bpy.context.evaluated_depsgraph_get()

    @property
    def fov(self):
        """Field of view in degrees"""
        return self.camera.data.angle * 180 / np.pi

    @fov.setter
    def fov(self, fov):
        self.set_fov(fov)

    # FOV is not animatable, so keyframe focal length instead
    @animatable_property("lens", "data")
    def set_fov(self, fov):
        self.camera.data.angle = fov * np.pi / 180
        self.update()

    @property
    def focal_length(self):
        """Focal length in mm"""
        return self.camera.data.lens

    @focal_length.setter
    def focal_length(self, focal_length):
        self.set_focal_length(focal_length)

    @animatable_property("lens", "data")
    def set_focal_length(self, focal_length):
        self.camera.data.lens = focal_length
        self.update()

    @property
    def clip_start(self):
        return self.camera.data.clip_start

    @clip_start.setter
    def clip_start(self, clip_start):
        self.set_clip_start(clip_start)

    @animatable_property("clip_start", "data")
    def set_clip_start(self, clip_start):
        self.camera.data.clip_start = clip_start
        self.update()

    @property
    def clip_end(self):
        return self.camera.data.clip_end

    @clip_end.setter
    def clip_end(self, clip_end):
        self.set_clip_end(clip_end)

    @animatable_property("clip_end", "data")
    def set_clip_end(self, clip_end):
        self.camera.data.clip_end = clip_end
        self.update()

    def look_at(
        self,
        at: types.VectorLike = mathutils.Vector((0, 0, 0)),
        up: types.VectorLike = mathutils.Vector((0, 0, 1)),
    ):
        """Look at a point in space, with up vector.

        :param at: Point to look at
        :param up: Up vector"""
        at = handle_vec(at, 3)
        up = handle_vec(up, 3)

        self.rotation_euler = _look_at_rotation(self.camera, at, up)

    def look_at_object(
        self,
        obj: Union[Mesh, bpy.types.Object],
        up: types.VectorLike = mathutils.Vector((0, 0, 1)),
    ):
        """Look at an object, with up vector.

        :param obj: Object to look at
        :param up: Up vector"""

        if isinstance(obj, Mesh):
            obj = obj.obj

        self.look_at(obj.location, up)

    def place_and_rotate(self, pos, euler):
        self.location = pos
        self.euler = euler

    def place_and_look_at(self, pos, at, up=mathutils.Vector((0, 0, 1))):
        pos = handle_vec(pos)
        at = handle_vec(at)
        up = handle_vec(up)

        self.location = pos
        self.look_at(at, up)



import os

from .blender_threading import BlenderThreadManager, _list_split
from ..utils.blender_setup.blender_locator import get_blender_path
from ..file.tempfiles import cleanup_temp_files as cleanup
from typing import Union

from sys import platform

if platform == "linux" or platform == "linux2" or platform == "darwin":
    nul_text = "/dev/null"
elif platform == "win32":
    nul_text = "nul"


class BlenderCommand:
    """Construct command for running blender script"""

    def __init__(self, blender_loc, blend_src=None, background=True):
        self.blender_loc = blender_loc
        self.blend_src = blend_src
        self.background = background
        self._command = [
            blender_loc,
            blend_src if blend_src else "",
            "--background" if background else "",
        ]

    def compose(self, script, args=(), **kwargs):
        command = self._command + ["--python", script]
        command += ["--"]
        for arg in args:
            command.append(f"--{arg}")

        for k, v in kwargs.items():
            command += [f"--{k}", str(v)]

        self._command = command

    @property
    def command(self):
        return self._command

    def copy(self):
        """Return a copy of the command"""
        cmd = BlenderCommand(self.blender_loc, self.background)
        cmd._command = [*self._command]
        return cmd

    def set_job(self, job_list):
        """Receives list of json files, adds as , separated string to command"""
        cmd = self.copy()
        cmd._command += ["--jobs", ",".join(job_list)]
        return cmd

    def set_logger(self, log_loc):
        """Set file for logging"""
        cmd = self.copy()
        cmd._command += ["--log", log_loc]
        return cmd


class Runner:
    def __init__(
        self,
        script,
        jsons,
        output_directory,
        num_threads=1,
        print_to_stdout=False,
        distributed: tuple = None,
        blend_src: str = None,
        **script_kwargs,
    ):
        """
        :param jsons: N sized list of .json files, each with info about the given job
        :param num_threads: threads to run in parallel (default = 1)
        :param distributed: tuple of (machine index, total machines) for distributed rendering
        :param blend_src: path to blend file to open (note: this is preferable to `blendersynth.load_blend` as it handles context better)
        """

        # if distributed, split jsons into chunks
        if distributed is not None:
            machine_index, total_machines = distributed
            assert (
                machine_index < total_machines
            ), "machine_index must be less than total_machines"
            jsons = _list_split(jsons, total_machines)[machine_index]

        self.jsons = jsons
        self.num_threads = num_threads

        # Split the jsons into num_threads chunks
        json_chunks = _list_split(self.jsons, self.num_threads)

        blender_loc = get_blender_path()
        command = BlenderCommand(blender_loc=blender_loc, background=True, blend_src=blend_src)
        command.compose(script=script, **script_kwargs)

        thread_manager = BlenderThreadManager(
            command,
            json_chunks,
            print_to_stdout=print_to_stdout,
            output_directory=output_directory,
            script_directory=os.path.dirname(script),
        )

        thread_manager.start()


def execute_jobs(
    script: str,
    json_src: Union[list, str],
    output_directory: str,
    num_threads: int = 1,
    print_to_stdout: bool = False,
    distributed: tuple = None,
    blend_src: str = None,
    **script_kwargs,
):
    """
    Execute a script, given a list of jobs to execute.

    :param script: path to script to execute
    :param json_src: N sized list of .json files, each with info about the given job. OR a directory containing .json files,
            which will be used as the jsons
    :param output_directory: directory to save output files
    :param num_threads: threads to run in parallel (default = 1)
    :param print_to_stdout: If True, print to stdout instead of saving to file
    :param distributed: tuple of (machine index, total machines) for distributed rendering
    :param blend_src: path to blend file to open (note: this is preferable to `blendersynth.load_blend` as it handles context better)
    """

    assert not (
        num_threads > 1 and print_to_stdout
    ), "Cannot print to stdout with multiple threads"
    if print_to_stdout:
        raise NotImplementedError("Print to stdout not supported yet.")

    if isinstance(json_src, str):
        json_src = sorted(
            [
                os.path.join(json_src, f)
                for f in os.listdir(json_src)
                if f.endswith(".json")
            ]
        )

    Runner(
        script,
        json_src,
        output_directory,
        num_threads,
        print_to_stdout=print_to_stdout,
        distributed=distributed,
        blend_src=blend_src,
        **script_kwargs,
    )

    cleanup()



import os.path

import bpy
from .utils import (
    GetNewObject,
    SelectObjects,
    handle_vec,
    CursorAt,
    animatable_property,
)
from .bsyn_object import BsynObject
from .material import Material
from .armature import Armature
from .aov import AOV
import numpy as np
import mathutils
import bmesh
from mathutils import Vector, Euler
from typing import Union, List
from copy import deepcopy
from ..utils import types

_primitives = {
    "cube": bpy.ops.mesh.primitive_cube_add,
    "sphere": bpy.ops.mesh.primitive_uv_sphere_add,
    "cylinder": bpy.ops.mesh.primitive_cylinder_add,
    "plane": bpy.ops.mesh.primitive_plane_add,
    "cone": bpy.ops.mesh.primitive_cone_add,
    "monkey": bpy.ops.mesh.primitive_monkey_add,
    "torus": bpy.ops.mesh.primitive_torus_add,
}

default_ids = {
    "loaded_mesh": 0,
    "prim_cube": 1,
    "prim_sphere": 2,
    "prim_cylinder": 3,
    "prim_plane": 4,
    "prim_cone": 5,
    "prim_monkey": 6,
    "prim_torus": 7,
}


def _get_child_meshes(obj):
    """Given an object, return all meshes that are children of it. Recursively searches children of children.
    Also returns any child objects that aren't meshes"""
    if obj.type == "MESH":
        return [obj], []
    else:
        meshes, other = [], [obj]
        for child in obj.children:
            child_meshes, child_other = _get_child_meshes(child)
            meshes += child_meshes
            other += child_other

        return meshes, other


def _set_object_origin(obj: bpy.types.Object, origin: Vector):
    """Set the origin of an object to a given point"""
    with SelectObjects([obj]), CursorAt(origin):
        bpy.ops.object.origin_set(type="ORIGIN_CURSOR", center="MEDIAN")


def _find_material(obj):
    """Return first material found in an object"""
    if obj.type == "MESH":
        for mat in obj.data.materials:
            return mat

    for child in obj.children:
        if child.type == "MESH":
            for mat in child.data.materials:
                return mat


class Mesh(BsynObject):
    """A mesh object. Can be a single mesh, or a hierarchy of meshes."""

    primitive_list = list(_primitives.keys())
    """List of available primitives"""

    def __init__(
        self,
        obj,
        material: bpy.types.Material = None,
        scene: bpy.types.Scene = None,
        class_id: int = None,
    ):
        """
        :param obj: Receives either a single mesh, or an empty with children empty & meshes
        :param material: bpy.types.Material to assign to the mesh
        :param scene:

        If obj contains multiple meshes, they will all be assigned the same material, and must act as a single object
        i.e. they will all be transformed together

        """

        if scene is None:
            scene = bpy.context.scene

        self.scene = scene
        self._object = obj
        self._meshes, self._other_objects = _get_child_meshes(obj)

        # Manage materials here
        if material is not None:
            self._material = Material.from_blender_material(material)
        elif _find_material(obj) is not None:
            self._material = Material.from_blender_material(_find_material(obj))
        else:
            self._material = Material("NewMaterial")

        for mesh in self._meshes:
            mesh.data.materials.append(self._material.object)

        # INSTANCING - Define InstanceID based on number of meshes in scene
        self._object["instance_id"] = scene.get("NUM_MESHES", 0)
        scene["NUM_MESHES"] = (
            scene.get("NUM_MESHES", 0) + 1
        )  # Increment number of meshes in scene

        # CLASS - Define class based on type of object (e.g. primitive)
        # can be overriden at any point
        self.set_class_id(class_id)

        self.assigned_aovs = []
        self._armatures = (
            {}
        )  # name : armature, to prevent multiple instances of armatures

    def set_class_id(self, class_id: int):
        """Set the class ID of the object.

        :param class_id: Class ID to assign to object
        """

        assert isinstance(
            class_id, int
        ), f"Class ID must be an integer, not {type(class_id)}"
        assert class_id >= 0, f"Class ID must be >= 0, not {class_id}"

        self._object["class_id"] = class_id
        self.scene["MAX_CLASSES"] = max(self.scene.get("MAX_CLASSES", 0), class_id)

    @classmethod
    def from_scene(cls, key: str, class_id: int = 0) -> "Mesh":
        """Create object from named object in scene.

        :param key: Name of object in scene
        :param class_id: Class ID to assign to object

        :return: Mesh loaded from scene"""
        obj = bpy.data.objects[key]
        return cls(obj, class_id=class_id)

    @classmethod
    def from_primitive(
        cls,
        name="cube",
        scale=None,
        location=None,
        rotation_euler=None,
        class_id=None,
        **kwargs,
    ) -> "Mesh":
        """Create Mesh from primitive.

        :param name: Name of primitive to create. See :attr:`~blendersynth.blender.Mesh.primitive_list` for options
        :param scale: Scale of object
        :param location: Location of object
        :param rotation_euler: Rotation of object
        :param class_id: Class ID to assign to object
        :param kwargs: Additional arguments to pass to primitive (see `bpy.ops.mesh.primitive_cube_add <https://docs.blender.org/api/current/bpy.ops.mesh.html#bpy.ops.mesh.primitive_cube_add>`_, etc.)

        :return: Mesh object created from primitive
        """

        assert (
            name in _primitives
        ), f"Primitive `{name}` not found. Options are: {list(_primitives.keys())}"

        importer = GetNewObject(bpy.context.scene)
        with importer:
            prim = _primitives[name](**kwargs)  # Create primitive

        if class_id is None:
            class_id = default_ids[f"prim_{name}"]

        obj = cls(importer.imported_obj, class_id=class_id)

        if scale is not None:
            obj.scale = scale
        if location is not None:
            obj.location = location
        if rotation_euler is not None:
            obj.rotation_euler = rotation_euler

        return obj

    @classmethod
    def from_numpy(
        cls, vertices: np.ndarray, faces: np.ndarray, name="New_Mesh"
    ) -> "Mesh":
        """Create Mesh from numpy arrays of vertices and faces."""
        mesh = bpy.data.meshes.new(name=name)
        obj = bpy.data.objects.new(name, mesh)
        bpy.context.collection.objects.link(obj)
        mesh.from_pydata(vertices.tolist(), [], faces.tolist())
        mesh.update()

        return cls(obj, class_id=default_ids["loaded_mesh"])

    @classmethod
    def from_obj(
        cls,
        obj_loc: str,
        class_id: int = None,
        forward_axis: str = "-Z",
        up_axis: str = "Y",
    ) -> "Mesh":
        """Load object from .obj file.

        :param obj_loc: Location of .obj file
        :param class_id: Class ID to assign to object
        :param forward_axis: Axis to use as forward axis
        :param up_axis: Axis to use as up axis


        Note: we use bpy.ops.wm.obj_import instead of bpy.ops.import_scene.obj because the latter
        causes issues with materials & vertex ordering.
        (Changing vertex ordering makes the use of keypoints difficult.)
        """
        for axis in (forward_axis, up_axis):
            assert axis in (
                "X",
                "Y",
                "Z",
                "-X",
                "-Y",
                "-Z",
            ), f"Axis `{axis}` not valid, must be one of X, Y, Z, -X, -Y, -Z"

        forward_axis = forward_axis.replace("-", "NEGATIVE_")  # e.g. -X -> NEGATIVE_X
        up_axis = up_axis.replace("-", "NEGATIVE_")

        assert os.path.isfile(obj_loc) and obj_loc.endswith(
            ".obj"
        ), f"File `{obj_loc}` not a valid .obj file"

        directory, fname = os.path.split(obj_loc)
        directory = os.path.abspath(directory)  # ensure absolute path for clean loading

        importer = GetNewObject(bpy.context.scene)
        with importer:
            bpy.ops.wm.obj_import(
                filepath=fname,
                directory=directory,
                filter_image=False,
                files=[{"name": fname}],
                forward_axis=forward_axis,
                up_axis=up_axis,
            )

        if class_id is None:
            class_id = default_ids["loaded_mesh"]

        obj = importer.imported_obj
        return cls(obj, class_id=class_id)

    @classmethod
    def from_glb(cls, glb_loc: str, class_id: int = None) -> "Mesh":
        """Load object from .glb file.

        :param glb_loc: Location of .glb file
        :param class_id: Class ID to assign to object
        """
        assert os.path.isfile(glb_loc) and glb_loc.endswith(
            (".glb", ".gtlf")
        ), f"File `{glb_loc}` not a valid .glb file"

        directory, fname = os.path.split(glb_loc)
        importer = GetNewObject(bpy.context.scene)
        with importer:
            bpy.ops.import_scene.gltf(filepath=glb_loc, files=[{"name": fname}])

        if class_id is None:
            class_id = default_ids["loaded_mesh"]

        return cls(importer.imported_obj, class_id=class_id)

    @classmethod
    def from_gltf(cls, gltf_loc: str, class_id: int = None) -> "Mesh":
        """Alias for :meth:`~blendersynth.blender.Mesh.from_glb`"""
        return cls.from_glb(gltf_loc, class_id=class_id)

    @classmethod
    def from_fbx(cls, fbx_loc: str, class_id: int = None) -> "Mesh":
        """Load object from .fbx file.

        :param fbx_loc: Location of .fbx file
        :param class_id: Class ID to assign to object"""
        assert os.path.isfile(fbx_loc) and fbx_loc.endswith(
            ".fbx"
        ), f"File `{fbx_loc}` not a valid .fbx file"

        directory, fname = os.path.split(fbx_loc)
        importer = GetNewObject(bpy.context.scene)
        with importer:
            bpy.ops.import_scene.fbx(filepath=fbx_loc, files=[{"name": fname}])

        if class_id is None:
            class_id = default_ids["loaded_mesh"]

        return cls(importer.imported_obj, class_id=class_id)

    @property
    def origin(self) -> Union[Vector, List[Vector]]:
        """
        Return origin of primary object.

        To get origins of all objects within the mesh, use :attr:`~blendersynth.blender.Mesh.all_origins`
        """
        return self.obj.location

    @origin.setter
    def origin(self, origin: Vector):
        """Set origin of primary object.

        :param origin: Origin to set

        To set origins of all objects within the mesh, use :attr:`~blendersynth.blender.Mesh.all_origins`
        """
        try:
            vec = handle_vec(origin)
        except ValueError:
            raise ValueError(
                f"Error with setting origin. Expects a 3 long Vector. Received: {origin}"
            )

        _set_object_origin(self._meshes[0], vec)

    @property
    def all_origins(self) -> List[Vector]:
        """Return list of origins of all objects within the mesh"""
        return [m.location for m in self._all_objects]

    @all_origins.setter
    def all_origins(self, origins: List[Vector]):
        """
        Set origins of all objects within the mesh.
        :param origins: List of origins to set
        """

        assert len(origins) == len(
            self._all_objects
        ), f"Expected {len(self._all_objects)} origins, got {len(origins)}"

        for i in range(len(self._all_objects)):
            try:
                vec = handle_vec(origins[i])
                _set_object_origin(self._all_objects[i], vec)
            except:
                raise ValueError(
                    f"Error with setting origins. Expects a list of {len(self._all_objects)} 3-long Vector. Received: {origins}"
                )

    def _get_all_vertices(self, ref_frame="WORLD") -> np.ndarray:
        """
        Get all vertices of object, taking into account deformations.
        :param ref_frame: LOCAL or WORLD
        :return: Nx3 array of vertices
        """

        depsgraph = bpy.context.evaluated_depsgraph_get()  # to account for deformations

        if ref_frame not in {"LOCAL", "WORLD"}:
            raise ValueError(
                f"Invalid ref_frame: {ref_frame}. Must be one of ['LOCAL', 'WORLD']"
            )

        verts = []

        for mesh in self._meshes:
            # use bmesh to get vertices - this accounts for deformations in depsgraph
            bm = bmesh.new()
            bm.from_object(mesh, depsgraph)
            bm.verts.ensure_lookup_table()
            mesh_verts = np.array([x.co for x in bm.verts])
            bm.free()

            if ref_frame == "WORLD":
                mesh_verts = np.dot(
                    mesh.matrix_world,
                    np.vstack((mesh_verts.T, np.ones(mesh_verts.shape[0]))),
                )

            verts.append(mesh_verts)

        verts = np.concatenate(verts, axis=1)
        verts /= verts[3]  # convert from homogeneous coordinates
        return verts[:3].T

    def get_raw_bounds(self) -> [Vector, Vector]:
        """Get the minimum and maximum bounds of all meshes in the object, under no deformations or transformations."""
        verts = np.array([v.co for mesh in self._meshes for v in mesh.data.vertices])
        bbox_min = Vector([*np.min(verts, axis=0)])
        bbox_max = Vector([*np.max(verts, axis=0)])
        return bbox_min, bbox_max

    @property
    def materials(self):
        return {m for mesh in self._meshes for m in mesh.data.materials}

    def assign_pass_index(self, index: int):
        """Assign pass index to object. This can be used when mask rendering.

        :param index: Pass index to assign"""
        for mesh in self._meshes:
            mesh.pass_index = index
        return index

    def assign_aov(self, aov: AOV):
        """Assign AOV to object. Applies to all materials.

        :param aov: AOV to assign"""
        if aov not in self.assigned_aovs:
            for material in self.materials:
                shader_node_tree = material.node_tree
                assert shader_node_tree is not None, "Material must have a node tree"
                aov.add_to_shader(shader_node_tree)

        self.assigned_aovs.append(aov)

    def assign_aovs(self, aovs: List[AOV]):
        """Assign multiple AOVs to object. Applies to all materials.

        :param aovs: AOVs to assign"""
        for aov in aovs:
            self.assign_aov(aov)

    def set_minimum_to(self, axis: str = "Z", pos: float = 0):
        """Set minimum of object to a given position in a given axis.

        :param axis: Axis to set minimum in
        :param pos: Position to set minimum to
        """
        min_pos = self._get_all_vertices("WORLD")[:, "XYZ".index(axis)].min()
        trans_vec = np.zeros(3)
        trans_vec["XYZ".index(axis)] = pos - min_pos
        self.translate(trans_vec)

    def clamp_in_axis(
        self, axis: str = "Z", mode: str = "min", value: float = 0
    ) -> np.ndarray:
        """Clamp object in a given axis to a given value - ensure that, in this axis,
        the object never goes below (mode='min') or above (mode='max') the given value.

        This can be used to ensure that an object never goes below the ground plane, for example.

        :param axis: Axis to clamp in
        :param mode: 'min' or 'max'
        :param value: Value to clamp to
        :return: Vector of translation applied to object
        """

        axis = axis.upper()
        assert axis in "XYZ", f"Invalid axis: {axis}"
        assert mode in {"min", "max"}, f"Invalid mode: {mode}"

        verts = self._get_all_vertices("WORLD")
        axis_idx = "XYZ".index(axis)

        translation = 0
        if mode == "min":
            translation = max(
                value - verts[:, axis_idx].min(), 0
            )  # can only be positive
        elif mode == "max":
            translation = min(
                value - verts[:, axis_idx].max(), 0
            )  # can only be negative

        trans_vec = np.zeros(3)
        trans_vec["XYZ".index(axis)] = translation
        self.translate(trans_vec)

        return trans_vec

    def delete(self, delete_materials: bool = True):
        """Clear mesh from scene & mesh data.

        :param delete_materials: Also delete object materials from scene"""
        mesh_names = [m.name for m in self._meshes]
        for mesh in self._meshes:
            bpy.data.objects.remove(mesh, do_unlink=True)

        # remove any non-mesh objects
        for obj in self._other_objects:
            bpy.data.objects.remove(obj, do_unlink=True)

        # Also remove its mesh data from bpy.data.meshes
        for mname in mesh_names:
            for m in bpy.data.meshes:
                if m.name == mname:
                    bpy.data.meshes.remove(m)
                    break

        self._meshes = []  # clear mesh list

        if delete_materials:
            for material in self.materials:
                bpy.data.materials.remove(material, do_unlink=True)

    def centroid(self, method: str = "median", seperate: bool = False) -> Vector:
        """
        Return the centroid of the mesh(es)

        :param method: See :attr:`~blendersynth.blender.mesh.Mesh.origin_to_centroid` for options.
        :param seperate: If True, will return a list of centroids for each object, rather than a single centroid.
        :return: Centroid of the mesh(es). If multiple meshes, will average the centroids.
        """

        if seperate:
            original_origins = deepcopy(self.all_origins)
            self.origin_to_centroid(method=method)
            centroids = deepcopy(self.all_origins)
            self.all_origins = original_origins
            return sum(centroids) / len(centroids)

        else:
            original_origin = deepcopy(self.origin)
            self.origin_to_centroid(method=method)
            centroid = deepcopy(self.origin)
            self.origin = original_origin
            return centroid

    def _set_origin_manual(self, origin: Vector, all_meshes=True):
        """Override to set origin manually. Should only be used by internal functions."""
        if all_meshes:
            for mesh in self._meshes:
                mesh.location = origin
        else:
            self._meshes[0].location = origin

    def origin_to_centroid(self, method: str = "bounds"):
        """Move object origin to centroid.

        Four methods available:

        * 'bounds' - move the origin to the center of the bounds of the mesh
        * 'median' - move the origin to the median point of the mesh
        * 'com_volume' - move to the centre of mass of the volume of the mesh
        * 'com_area' - Move to the centre of mass of the surface of the mesh

        :param method: Selected method to move origin to centroid
        """

        _valid_methods = ["bounds", "median", "com_volume", "com_area"]

        _type_lookup = dict(
            bounds="ORIGIN_GEOMETRY",
            median="ORIGIN_GEOMETRY",
            com_volume="ORIGIN_CENTER_OF_VOLUME",
            com_area="ORIGIN_CENTER_OF_MASS",
        )

        center = "BOUNDS" if method == "bounds" else "MEDIAN"

        with SelectObjects(self._meshes + self._other_objects):
            bpy.ops.object.origin_set(type=_type_lookup[method], center=center)

    def get_keypoints(
        self, idxs: list = None, position: Union[np.ndarray, List[Vector]] = None
    ) -> List[Vector]:
        """Return 3D keypoint positions in world coordinates, given either:

        :param idxs: list of indices or ndarray of keypoints to project (only valid for single-mesh objects)
        :param position: 3D position of keypoints to project - in LOCAL object coordinates

        :return: N list of Vectors of keypoints in world space, where N is the number of keypoints
        """

        assert (idxs is not None) ^ (
            position is not None
        ), "Must provide either idxs or position, but not both."

        if idxs is not None:
            assert (
                len(self._meshes) == 1
            ), "Can only project keypoints by index for single-mesh objects."
            kps3d = [
                self.matrix_world @ self._meshes[0].data.vertices[i].co for i in idxs
            ]

        elif position is not None:
            kps3d = [self.matrix_world @ p for p in position]

        return kps3d

    @property
    def name(self):
        return self._meshes[0].name

    def get_armature(self, armature_name: str = None) -> Armature:
        """Get armature.
        If no name given, return first armature found.

        :param armature_name: Name of armature to load."""

        armatures = [obj for obj in self._other_objects if obj.type == "ARMATURE"]

        if len(armatures) == 0:
            raise ValueError("No armatures found.")

        if armature_name is None:
            armature_name = armatures[0].name

        if armature_name in self._armatures:
            return self._armatures[armature_name]

        for arm in armatures:
            if arm.name == armature_name:
                bsyn_arm = Armature(arm)
                self._armatures[armature_name] = bsyn_arm
                return bsyn_arm

        raise KeyError(f"Armature `{armature_name}` not found.")

    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, material: Material):
        self._material = material
        for mesh in self._meshes:
            mesh.data.materials.clear()
            mesh.data.materials.append(material.object)

    @property
    def shape_keys(self) -> bpy.types.Key:
        """Get the first shape keys object available"""
        for mesh in self._meshes:
            if mesh.data.shape_keys is not None:
                return mesh.data.shape_keys

        return None

    def get_shape_key(self, name: str, create: bool = False) -> bpy.types.ShapeKey:
        """Get a given shape key object.

        :param name: Name of shape key
        :param create: If True, create shape key if it doesn't exist. Otherwise, raise error.
        """
        keys = self.shape_keys
        if keys is None:
            if create:
                self.obj.shape_key_add(name="Basis")
                keys = self.shape_keys
            else:
                raise KeyError(f"Tried to get key '{name}' - no shape keys available.")

        if name not in keys.key_blocks:
            if create:
                self.create_shape_key(name)
            else:
                raise KeyError(f"Tried to get key '{name}' - not found in shape keys.")

        return keys.key_blocks[name]

    def create_shape_key(self, name: str):
        """Create a new shape key with a given name."""
        self.obj.shape_key_add(name=name)

    def set_shape_key(self, name: str, value: float, frame: int = None):
        """Set shape key to a given value.

        :param name: Name of shape key
        :param value: Value to set shape key to
        :param frame: If not None, set keyframe at this frame
        """

        shape_key = self.get_shape_key(name)
        shape_key.value = value

        if frame is not None:
            shape_key.keyframe_insert(data_path="value", frame=frame)

    def set_shape_keys(self, data: dict, frame: int = None):
        """Set multiple shape keys at once.

        :param data: Dictionary of shape key names to values
        :param frame: If not None, set keyframe at this frame
        """
        for k, v in data.items():
            self.set_shape_key(k, v, frame=frame)

    def set_shape_key_data(
        self,
        name: str,
        data: np.ndarray,
        create: bool = True,
        value_min: float = 0,
        value_max: float = 1,
    ):
        """Set shape key to a given value.

        :param name: Name of shape key
        :param data: Data to set shape key to (absolute positions - *not* offsets)
        :param create: If True, create shape key if it doesn't exist. Otherwise, raise error.
        :param value_min: Minimum value of shape key
        :param value_max: Maximum value of shape key
        """

        shape_key = self.get_shape_key(name, create=create)
        for i, coord in enumerate(data):
            shape_key.data[i].co = coord

        shape_key.slider_min = value_min
        shape_key.slider_max = value_max

    def make_shape_key(
        self, name: str, data: np.ndarray, value_min: float = 0, value_max: float = 1
    ):
        """Create a new shape key, optionally with data.

        :param name: Name of shape key
        :param data: Data to set shape key to
        :param value_min: Minimum value of shape key
        :param value_max: Maximum value of shape key
        """

        if self.shape_keys is None:
            self.obj.shape_key_add(name="Basis")  # add basis shape key

        if name in self.shape_keys.key_blocks:
            raise ValueError(f"Shape key `{name}` already exists.")

        self.obj.shape_key_add(name=name)
        if data is not None:
            self.set_shape_key_data(
                name, data, value_min=value_min, value_max=value_max
            )

    @property
    def _all_objects(self):
        """List of all objects associated with this object."""
        return self._meshes + self._other_objects

    def add_child(self, obj: Union["Mesh", bpy.types.Object]):
        """Add child Mesh or Object to this object.
        This will make the child object move with this object.

        :param obj: Object to add as child
        """

        new_objects = []
        if isinstance(obj, Mesh):
            self._meshes += obj._meshes
            self._other_objects += obj._other_objects
            new_objects = obj._all_objects

        else:
            if obj.type == "MESH":
                self._meshes.append(obj)
            else:
                self._other_objects.append(obj)

            new_objects.append(obj)

        # add as children
        with SelectObjects(new_objects, active_object=self.obj):
            bpy.ops.object.parent_set(type="OBJECT", keep_transform=True)

    def join(self, other_meshes: Union[List["Mesh"], List[bpy.types.Object]]):
        """
        Join other meshes into this mesh.
        :param other_meshes: List of either Mesh or blender Object types
        :return:
        """

        objects_to_join = []

        for other in other_meshes:
            if isinstance(other, Mesh):
                objects_to_join += other._all_objects

            elif isinstance(other, bpy.types.Object):
                objects_to_join += other

            else:
                raise ValueError(f"Cannot join object of type {type(other)} into Mesh.")

        with SelectObjects(objects_to_join, active_object=self.obj):
            bpy.ops.object.join()

    def export_obj(self, obj_loc: str, **kwargs):
        """Export object to .obj file.

        :param obj_loc: Location to save .obj file to
        :param kwargs: Additional arguments to pass to `bpy.ops.wm.obj_export`
        """
        with SelectObjects(self._all_objects):
            bpy.ops.wm.obj_export(filepath=obj_loc,
                                  export_selected_objects=True, **kwargs)



import ffmpeg
import os


def hstack(
    videos: list, out_path: str, overwrite: bool = True, loglevel: str = "error"
):
    """Horizontally stack videos.

    :param videos: list of video paths
    :param out_path: output video path
    :param overwrite: overwrite existing video file
    :param loglevel: ffmpeg loglevel (see ffmpeg documentation, default error)"""

    (
        ffmpeg.filter([ffmpeg.input(v) for v in videos], "hstack")
        .output(out_path)
        .global_args("-loglevel", loglevel)
        .run(overwrite_output=overwrite)
    )



"""Shader AOV manager"""

import bpy
from .nodes import tidy_tree, DeformedGeneratedTextureCoordinates, tree_add_socket
from ..utils import types
from typing import Union

ref_frames = ["CAMERA", "WORLD", "OBJECT"]

# Acceptable socket types for AOV colors & nodes
_socket_color_types = (bpy.types.NodeSocketVector, bpy.types.NodeSocketColor)
_socket_value_types = (bpy.types.NodeSocketFloat, bpy.types.NodeSocketInt)


# docs-special-members: __init__
# no-inherited-members


class AOV:
    """A generic Arbitrary Output Value.
    An AOV is a float or color value per-point that is calculated at the rendering stage,
    and can be output in a compositor to form an image.
    See `Blender docs <https://docs.blender.org/manual/en/latest/render/shader_nodes/output/aov.html>`_ for more info.
    """

    def __init__(self, *, name=None, **kwargs):
        if name is None:
            name = self.__class__.__name__

        self.name = name
        self._aov = None
        self._add_to_view_layer()

    def _add_to_view_layer(self, view_layer=None):
        """Add aov to view layer if not already there"""
        if view_layer is None:
            view_layer = bpy.context.view_layer

        for aov in view_layer.aovs:
            if aov.name == self.name:
                return

        aov = view_layer.aovs.add()
        aov.name = self.name
        self._aov = aov

    def add_to_shader(self, shader_node_tree):
        """Add AOV to shader node tree"""
        # add_to_shader must return output socket to connect to node
        out_socket = self._add_to_shader(shader_node_tree)

        # check node socket is correct type
        AOV_TYPE = None
        if isinstance(out_socket, _socket_color_types):
            AOV_TYPE = "COLOR"
        elif isinstance(out_socket, _socket_value_types):
            AOV_TYPE = "VALUE"
        else:
            raise ValueError(
                f"Output of _add_to_layer must be in {_socket_color_types} if Color or {_socket_value_types} if value. Got: `{type(out_socket)}`"
            )

        shader_aov_node = shader_node_tree.nodes.new("ShaderNodeOutputAOV")
        shader_aov_node.name = self.name
        shader_node_tree.links.new(out_socket, shader_aov_node.inputs[AOV_TYPE.title()])

        self._aov.type = AOV_TYPE
        tidy_tree(shader_node_tree)

    def _add_to_shader(self, shader_node_tree) -> bpy.types.NodeSocket:
        raise NotImplementedError

    def update(self):
        """Some AOVs need an update before rendering (to change certain node properties)"""
        return

    def __str__(self):
        return self.name


class NormalsAOV(AOV):
    def __init__(
        self,
        *,
        name: str = None,
        ref_frame: str = "CAMERA",
        order: str = "XYZ",
        polarity: Union[tuple, list] = (1, 1, 1),
    ):
        """Given a shader node tree, add surface normals as output.

        :param name: Name of AOV to add
        :param ref_frame: Reference frame to use for normals
        :param order: Order of components in RGB (default: XYZ)
        :param polarity: Polarity of XYZ (1 or -1)
        """
        super().__init__(name=name)
        assert ref_frame in ref_frames, f"ref_frame must be one of {ref_frames}"

        self.ref_frame = ref_frame
        self.order = order
        self.polarity = polarity

    def _add_to_shader(self, shader_node_tree):
        geom_node = shader_node_tree.nodes.new("ShaderNodeNewGeometry")
        vec_transform = shader_node_tree.nodes.new("ShaderNodeVectorTransform")
        map_range_node = shader_node_tree.nodes.new("ShaderNodeMapRange")

        vec_transform.vector_type = "NORMAL"
        vec_transform.convert_to = self.ref_frame

        # Set up mapping - with polarity
        map_range_node.data_type = "FLOAT_VECTOR"
        for i in range(3):
            map_range_node.inputs[7].default_value[i] = -self.polarity[i]
            map_range_node.inputs[8].default_value[i] = self.polarity[i]

        # Set up ordering
        sep_xyz_node = shader_node_tree.nodes.new("ShaderNodeSeparateXYZ")
        comb_xyz_node = shader_node_tree.nodes.new("ShaderNodeCombineXYZ")
        for i in range(3):
            shader_node_tree.links.new(
                sep_xyz_node.outputs[self.order[i]], comb_xyz_node.inputs["XYZ"[i]]
            )

        # Make necessary connections for shader graph
        shader_node_tree.links.new(
            geom_node.outputs["True Normal"], vec_transform.inputs["Vector"]
        )
        shader_node_tree.links.new(
            vec_transform.outputs["Vector"], map_range_node.inputs["Vector"]
        )
        shader_node_tree.links.new(
            map_range_node.outputs["Vector"], sep_xyz_node.inputs["Vector"]
        )

        return comb_xyz_node.outputs["Vector"]


class GeneratedAOV(AOV):
    """'Generated Texture Coordinates'.

    These are coordinates normalized to 0-1 for the object's undeformed bounding box, not taking into account
    deformation (pose, modifiers). See `Blender docs <https://docs.blender.org/manual/en/latest/render/shader_nodes/input/texture_coordinate.html>`_ for more info.
    """

    def _add_to_shader(self, shader_node_tree):
        texcon_node = shader_node_tree.nodes.new("ShaderNodeTexCoord")
        return texcon_node.outputs["Generated"]


class DisplacementGeneratedAOV(AOV):
    """In the same co-ordinate space as :class:`GeneratedAOV`, give the displacement vector
    for each point under modifiers (e.g. pose).

    By default, this deformation is mapped from the range [-1 to 1] to [0 to 1], with 0.5 representing no
    deformation. Any values outside of this range are clamped.
    This can be modified through the input kwargs vmin and vmax.
    """

    def __init__(
        self,
        *,
        name: str = None,
        mesh: types.Mesh = None,
        bbox_min: types.VectorLike = None,
        bbox_max: types.VectorLike = None,
        vmin: float = -1,
        vmax: float = 1,
    ):
        """Create AOV for displacement under modifiers.

        :param name: Name of AOV
        :param mesh: Mesh to calculate bounds for. Will be used to find bbox_min and bbox_max if given
        :param bbox_min: Minimum of bounding box. If not given, calculated from Mesh
        :param bbox_max: Maximum of bounding box. If not given, calculated from Mesh
        :param vmin: Minimum deformation to map to 0
        :param vmax: Maximum value to map to 1
        """

        super().__init__(name=name)
        self.mesh = mesh
        self.bbox_min = bbox_min
        self.bbox_max = bbox_max
        self.vmin = vmin
        self.vmax = vmax

        assert mesh is not None or (
            bbox_min is not None and bbox_max is not None
        ), "Either mesh or bbox_min and bbox_max must be given for DisplacementGeneratedAOV"

    def _add_to_shader(self, shader_node_tree):
        self.deformed_coords = DeformedGeneratedTextureCoordinates(
            node_tree=shader_node_tree,
            mesh=self.mesh,
            bbox_min=self.bbox_min,
            bbox_max=self.bbox_max,
        )

        self.generated_coords = shader_node_tree.nodes.new("ShaderNodeTexCoord")

        # subtract one from the other
        sub_node = shader_node_tree.nodes.new("ShaderNodeVectorMath")
        sub_node.operation = "SUBTRACT"

        # map result to range 0-1
        map_range_node = shader_node_tree.nodes.new("ShaderNodeMapRange")
        map_range_node.data_type = "FLOAT_VECTOR"
        map_range_node.inputs[7].default_value = [self.vmin] * 3
        map_range_node.inputs[8].default_value = [self.vmax] * 3

        # link up nodes
        shader_node_tree.links.new(
            self.deformed_coords.outputs["Vector"], sub_node.inputs[0]
        )
        shader_node_tree.links.new(
            self.generated_coords.outputs["Generated"], sub_node.inputs[1]
        )
        shader_node_tree.links.new(
            sub_node.outputs["Vector"], map_range_node.inputs["Vector"]
        )

        tidy_tree(shader_node_tree)
        return map_range_node.outputs["Vector"]

    def set_bounds(
        self,
        mesh: types.Mesh = None,
        bbox_min: types.VectorLike = None,
        bbox_max: types.VectorLike = None,
    ):
        """Set bounds for DeformedGeneratedTextureCoordinates node group

        :param mesh: Mesh to calculate bounds for. Will be used to find bbox_min and bbox_max if given
        :param bbox_min: Minimum of bounding box. If not given, calculated from Mesh
        :param bbox_max: Maximum of bounding box. If not given, calculated from Mesh
        """
        self.mesh = mesh
        self.bbox_min = bbox_min
        self.bbox_max = bbox_max
        self.node_group.register_bounds(mesh, bbox_min, bbox_max)


class UVAOV(AOV):
    """UV coordinates. See `Blender docs <https://docs.blender.org/manual/en/latest/editors/uv/index.html#editors-uv-index>`_ for more info."""

    def _add_to_shader(self, shader_node_tree):
        texcon_node = shader_node_tree.nodes.new("ShaderNodeTexCoord")
        return texcon_node.outputs["UV"]


class ValueAOV(AOV):
    """A generic AOV that outputs a single value"""

    def __init__(self, value: float = 0, name=None, **kwargs):
        """
        :param value: Value to output
        :param name: Name of AOV
        :param kwargs: Additional kwargs to pass to :class:`AOV`
        """

        super().__init__(name=name, **kwargs)
        self._value = 0
        self.value_node = None

        self.value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        if self.value_node is not None:
            self.value_node.outputs["Value"].default_value = value

    def _add_to_shader(self, shader_node_tree):
        self.value_node = shader_node_tree.nodes.new("ShaderNodeValue")
        self.value = self._value
        return self.value_node.outputs["Value"]


class AttrAOV(AOV):
    """Take an object attribute, and output it as an AOV"""

    attribute_type = None
    attribute_name = None

    def _add_to_shader(self, shader_node_tree):
        attr_node = shader_node_tree.nodes.new("ShaderNodeAttribute")
        attr_node.attribute_type = self.attribute_type
        attr_node.attribute_name = self.attribute_name
        return attr_node.outputs["Fac"]


class InstanceIDAOV(AttrAOV):
    """Instance ID - given to each object on creation.
    Output is an integer corresponding to the object's instance ID (0-indexed)
    """

    attribute_type = "OBJECT"
    attribute_name = "instance_id"


class ClassIDAOV(AttrAOV):
    """Class ID - given to each object on creation.
    Output is an integer corresponding to the object's class ID (0-indexed)
    Class IDs can be manually set either when creating a Mesh, or by using the Mesh's :meth:`~blendersynth.blender.mesh.Mesh.set_class_id` method.
    If not set, will default to a different index from each primitive.
    """

    attribute_type = "OBJECT"
    attribute_name = "class_id"


class AttrRGBAOV(AOV):
    """
    For a given numerical attribute, outputs a color corresponding to the attribute's value.
    Object with attribute value `i` has HSV color `(i/N, 1, 1)`.
    `N` can be a property of the object to update.
    Runs :meth:`~update()` method to change the value of `N`, which is called before rendering.
    """

    attribute_type = None
    attribute_name = None

    def __init__(self, *, name=None):
        super().__init__(name=name)

        # Create Int Index -> HSV as a node group, so the 'num_objects' parameter can be edited centrally
        self.group = bpy.data.node_groups.new(name="IdxToHue", type="ShaderNodeTree")

        tree_add_socket(self.group, "NodeSocketFloat", "Index", "INPUT")
        tree_add_socket(self.group, "NodeSocketColor", "Color", "OUTPUT")

        self.input_node = self.group.nodes.new("NodeGroupInput")
        self.output_node = self.group.nodes.new("NodeGroupOutput")

        self.div_node = div_node = self.group.nodes.new(
            "ShaderNodeMath"
        )  # Need to keep reference so can update at runtime
        div_node.operation = "DIVIDE"
        div_node.use_clamp = True
        div_node.inputs[1].default_value = self.N

        hsv_node = self.group.nodes.new("ShaderNodeHueSaturation")
        hsv_node.inputs["Saturation"].default_value = 1
        hsv_node.inputs["Value"].default_value = 1
        hsv_node.inputs["Color"].default_value = (1, 0, 0, 1)  # Red

        self.group.links.new(self.input_node.outputs["Index"], div_node.inputs[0])
        self.group.links.new(div_node.outputs["Value"], hsv_node.inputs["Hue"])
        self.group.links.new(
            hsv_node.outputs["Color"], self.output_node.inputs["Color"]
        )
        tidy_tree(self.group)

    def _add_to_shader(self, shader_node_tree):
        attr_node = shader_node_tree.nodes.new("ShaderNodeAttribute")
        attr_node.attribute_type = self.attribute_type
        attr_node.attribute_name = self.attribute_name

        # Create a group node for the Int Index -> HSV conversion
        group_node = shader_node_tree.nodes.new("ShaderNodeGroup")
        group_node.node_tree = self.group

        shader_node_tree.links.new(attr_node.outputs["Fac"], group_node.inputs["Index"])
        return group_node.outputs["Color"]

    def update(self):
        self.div_node.inputs[1].default_value = self.N

    @property
    def N(self):
        return 0


class InstanceRGBAOV(AttrRGBAOV):
    """
    :class:`~InstanceIDAOV` as an :class:`~AttrRGBAOV` for visualisation.
    Updates `N` at render time by reading the scene property 'NUM_MESHES'
    """

    attribute_type = "OBJECT"
    attribute_name = "instance_id"

    @property
    def N(self):
        """Total number of meshes in the scene"""
        return bpy.context.scene.get("NUM_MESHES", 0) + 1


class ClassRGBAOV(AttrRGBAOV):
    """
    :class:`~ClassIDAOV` but as an :class:`~AttrRGBAOV` for visualisation.
    Updates `N` at render time by reading the scene property `MAX_CLASSES`
    """

    attribute_type = "OBJECT"
    attribute_name = "class_id"

    @property
    def N(self):
        """Update the divisor node with the current number of classes"""
        return bpy.context.scene.get("MAX_CLASSES", 0) + 1



"""Base class for all BlenderSynth objects."""
from .utils import (
    handle_vec,
    SelectObjects,
    _euler_add,
    animatable_property,
    _euler_invert,
)
from mathutils import Vector, Euler, Matrix
import numpy as np
import bpy
from typing import Union
from ..utils import types

from typing import TYPE_CHECKING

# for documentation, import these types
if TYPE_CHECKING:
    from .curve import Curve


class BsynObject:
    """Generic class for BlenderSynth objects.

    Assigned an .obj (eg bpy.types.Mesh for a Mesh) which is the main Blender object it represents.
    """

    _object: bpy.types.Object = None  # corresponding blender object

    @property
    def obj(self) -> bpy.types.Object:
        if self._object is None:
            raise ValueError(
                "self._object not set. Ensure it is set in the Object's __init__ function."
            )

        return self._object

    @property
    def object(self) -> bpy.types.Object:
        return self.obj

    @property
    def data(self) -> bpy.types.ID:
        return self.object.data

    def remove(self):
        """Delete .obj from bpy.data if it exists"""
        try:
            bpy.data.objects.remove(self.obj)
        except ReferenceError:
            pass

    def update(self):
        #   ---> not currently needed, may be needed in the future
        # 	"""On any update, run this. For most objects, this is a no-op, but for some objects,
        # 	this is necessary to update the object's state. e.g. Camera"""
        return

    def _keyframe_delete(self, *args, **kwargs):
        self._object.keyframe_delete(*args, **kwargs)

    def _keyframe_insert(self, *args, **kwargs):
        self._object.keyframe_insert(*args, **kwargs)

    @property
    def name(self):
        return self.obj.name

    @property
    def _all_objects(self):
        """List of all objects associated with this object."""
        return [self.object]

    @property
    def origin(self) -> Vector:
        return self.location

    @property
    def location(self) -> Vector:
        """Location of object"""
        return self.obj.location

    @location.setter
    def location(self, value):
        self.set_location(value)

    @property
    def rotation_euler(self) -> Euler:
        """Rotation in euler XYZ angles"""
        return self.obj.rotation_euler

    @rotation_euler.setter
    def rotation_euler(self, value):
        self.set_rotation_euler(value)

    @property
    def scale(self):
        return self.obj.scale

    @scale.setter
    def scale(self, scale):
        self.set_scale(scale)

    @property
    def dimensions(self):
        return self.obj.dimensions

    @dimensions.setter
    def dimensions(self, dimensions):
        self.set_dimensions(dimensions)

    @animatable_property("location")
    def set_location(self, location: types.VectorLike):
        """Set location of object.

        :param location: Location vector to set"""
        location = handle_vec(location, 3)

        translation = location - self.location
        with SelectObjects(self._all_objects):
            bpy.ops.transform.translate(value=translation)

    def _apply_rotation(self, rot):
        for ax, val in zip("XYZ", rot):
            if val != 0:
                bpy.ops.transform.rotate(
                    value=val,
                    orient_axis=ax,
                    orient_type="GLOBAL",
                    constraint_axis=[ax == "X", ax == "Y", ax == "Z"],
                )

    @animatable_property("rotation_euler")
    def set_rotation_euler(self, rotation: types.VectorLike):
        """Set euler rotation of object.

        :param rotation: Rotation vector"""

        assert (
            len(rotation) == 3
        ), f"Rotation must be a tuple of length 3, got {len(rotation)}"
        rotation = Euler(rotation, "XYZ")

        # to avoid dealing with rotation calculations, we first rotate to the origin, then rotate to the new rotation
        with SelectObjects(self._all_objects):
            self._apply_rotation(
                _euler_invert(self.rotation_euler)
            )  # invert current rotation
            self._apply_rotation(rotation)  # apply new rotation

    @animatable_property("scale")
    def set_scale(self, scale: types.VectorLikeOrScalar):
        """Set scale of object.

        :param scale: Scale to set. Either single value or 3 long vector"""
        if isinstance(scale, (int, float)):
            scale = (scale, scale, scale)

        resize_fac = np.array(scale) / np.array(self.scale)

        with SelectObjects(self._all_objects):
            bpy.ops.transform.resize(value=resize_fac)

    @animatable_property("dimensions")
    def set_dimensions(self, dimensions: types.VectorLikeOrScalar):
        """Set dimensions of object.

        :param dimensions: Dimensions to set. Either single value or 3 long vector"""
        if isinstance(dimensions, (int, float)):
            dimensions = (dimensions, dimensions, dimensions)

        resize_fac = np.array(dimensions) / np.array(self.dimensions)

        with SelectObjects(self._all_objects):
            bpy.ops.transform.resize(value=resize_fac)

    def translate(self, translation):
        """Translate object"""
        translation = handle_vec(translation, 3)
        self.location = self.location + translation

    def rotate_by(self, rotation):
        """Add a rotation to the object. Must be in XYZ order, euler angles, radians."""
        rotation = handle_vec(rotation, 3)
        new_rotation = _euler_add(self.rotation_euler, Euler(rotation, "XYZ"))
        self.rotation_euler = new_rotation

    def scale_by(self, scale):
        """Scale object"""
        if isinstance(scale, (int, float)):
            scale = (scale, scale, scale)

        scale = handle_vec(scale, 3)
        self.scale *= scale

    @property
    def matrix_world(self):
        """Return world matrix of object(s)."""
        bpy.context.evaluated_depsgraph_get()  # required to update object matrix
        return self.object.matrix_world

    @property
    def axes(self) -> np.ndarray:
        """Return 3x3 rotation matrix (normalized) to represent axes"""
        mat = np.array(self.matrix_world)[:3, :3]
        mat = mat / np.linalg.norm(mat, axis=0)
        return mat

    def track_to(self, obj: Union["BsynObject", bpy.types.Object]):
        """Track to object.

        :param obj: BsynObject or Blender Object to track to
        """

        if isinstance(obj, BsynObject):
            obj = obj.obj

        constraint = self.object.constraints.new("TRACK_TO")
        constraint.target = obj
        constraint.track_axis = "TRACK_NEGATIVE_Z"
        constraint.up_axis = "UP_Y"

    def untrack(self):
        """Remove track to constraint from object"""
        constraint = self.object.constraints.get("Track To")
        self.object.constraints.remove(constraint)

    def follow_path(
        self,
        path: "Curve",
        zero: bool = True,
        animate: bool = True,
        frames: tuple = (0, 250),
        fracs: tuple = (0, 1),
    ):
        """Follow path, with optional animation setting.

        :param path: Curve object
        :param zero: If True, set camera location to (0, 0, 0) [aligns camera with path]
        :param animate: If True, animate camera along path
        :param frames: tuple of keyframes to animate at - length N
        :param fracs: tuple of fractions along path to animate at - length N
        """

        constraint = self.object.constraints.new("FOLLOW_PATH")
        constraint.target = path.path
        constraint.forward_axis = "TRACK_NEGATIVE_Z"
        constraint.up_axis = "UP_Y"
        constraint.use_fixed_location = (
            True  # ensures that offset factor is in 0-1 range
        )

        if zero:
            self.location = (0, 0, 0)

        if animate:
            self.animate_path(frames, fracs)

        # if there are any track constraints, place this constraint first
        # so that the object is not rotated by the track constraint
        track_constraint_idx = self.object.constraints.find("Track To")
        if track_constraint_idx > -1:
            self.object.constraints.move(track_constraint_idx, track_constraint_idx + 1)

    def animate_path(self, frames: tuple = (0, 250), fracs: tuple = (0, 1)):
        """Animate object along path.

        :param frames: tuple of keyframes to animate at - length N
        :param fracs: tuple of fractions along path to animate at - length N
        """

        assert len(frames) == len(
            fracs
        ), f"frames and fracs must be same length - got {len(frames)} and {len(fracs)}"

        for frame, frac in zip(frames, fracs):
            self.path_keyframe(frame, frac)

    def path_keyframe(self, frame: int, offset: float):
        """Set keyframe for camera path offset

        :param frame: Frame number
        :param offset: Offset fraction (0-1)
        """
        constraint = self.object.constraints.get("Follow Path")
        if constraint is None:
            raise ValueError(
                f"Object `{self.__class__}` does not have a 'Follow Path' constraint"
            )

        constraint.offset_factor = offset
        constraint.keyframe_insert(data_path="offset_factor", frame=frame)

    @animatable_property("hide_viewport")
    def viewport_visibility(self, value: bool):
        """Show/hide object in viewport

        :param value: True to show, False to hide
        """

        self.object.hide_viewport = not value

    @animatable_property("hide_render")
    def render_visibility(self, value: bool):
        """Show/hide object in render

        :param value: True to show, False to hide
        """
        self.object.hide_render = not value



import bpy


def is_version(version_num: int) -> bool:
    """Check if the Blender version is equal to the given version number.

    :param version_num: version number to check against
    """
    return int(bpy.app.version_string.split(".")[0]) == version_num


def is_version_plus(version_num: int) -> bool:
    """Check if the Blender version is greater than or equal to the given version number.

    is_version_plus(4.0) equivalent to is version 4.0+

    :param version_num: version number to check against
    """
    return int(bpy.app.version_string.split(".")[0]) >= version_num



# # Add mask node & Mix node
# mask_node = self.get_mask(mask_index)
# mix_node = self.node_tree.nodes.new('CompositorNodeMixRGB')
# mix_node.inputs[1].default_value = (0, 0, 0, 0)  # black - 0 alpha
#
# self.node_tree.links.new(mask_node.outputs['Alpha'], mix_node.inputs['Fac'])
# self.node_tree.links.new(self.render_layers_node.outputs[input_name], mix_node.inputs[2])
# self.node_tree.links.new(mix_node.outputs['Image'], node.inputs['Image'])

from ..nodes import CompositorNodeGroup


class MaskOverlay(CompositorNodeGroup):
    def __init__(self, name, node_tree, index=0, use_antialiasing=True, dtype="Color"):
        """
        Takes an IndexOB and an Image, and sets the values of all pixels with index != IndexOB
        to (0, 0, 0, 0)
        """

        super().__init__(name, node_tree)

        # define I/O
        self.add_socket("NodeSocketFloat", "IndexOB", "INPUT")
        self.add_socket(f"NodeSocket{dtype}", "Image", "INPUT")
        self.add_socket(f"NodeSocket{dtype}", "Image", "OUTPUT")

        # create nodes
        self.mask_node = self.group.nodes.new("CompositorNodeIDMask")
        self.mask_node.index = index
        self.mask_node.use_antialiasing = use_antialiasing

        self.mix_node = self.group.nodes.new("CompositorNodeMixRGB")
        self.mix_node.inputs[1].default_value = (0, 0, 0, 1)  # to set color to black

        self.alpha_node = self.group.nodes.new(
            "CompositorNodeSetAlpha"
        )  # to set alpha to 0
        self.alpha_node.mode = "APPLY"

        # link up internal nodes
        self.group.links.new(
            self.input_node.outputs["IndexOB"], self.mask_node.inputs["ID value"]
        )
        self.group.links.new(self.input_node.outputs["Image"], self.mix_node.inputs[2])
        self.group.links.new(
            self.mask_node.outputs["Alpha"], self.mix_node.inputs["Fac"]
        )
        self.group.links.new(
            self.mask_node.outputs["Alpha"], self.alpha_node.inputs["Alpha"]
        )
        self.group.links.new(
            self.mix_node.outputs["Image"], self.alpha_node.inputs["Image"]
        )
        self.group.links.new(
            self.alpha_node.outputs["Image"], self.output_node.inputs["Image"]
        )



"""Node Group collections for use in Blender"""

import bpy
from ...utils import types
import numpy as np
from mathutils import Vector
from .node_group import ShaderNodeGroup

# docs-special-members: __init__
# no-inherited-members


class DeformedGeneratedTextureCoordinates(ShaderNodeGroup):
    """
    A :class:`ShaderNodeGroup` that outputs generated texture coordinates, but in the deformed space.

    For example use, see :class:`DisplacementGeneratedAOV <blendersynth.blender.aov.DisplacementGeneratedAOV>`.
    """

    def __init__(
        self,
        node_tree: bpy.types.NodeTree,
        mesh: types.Mesh,
        bbox_min: types.VectorLike,
        bbox_max: types.VectorLike,
    ):
        """To calculate the generated space, either needs mesh or bbox_min, bbox_max.

        :param node_tree: NodeTree to add group to
        :param mesh: If given, `bbox_min` and `bbox_max` are calculated using this
        :param bbox_min: If given, `bbox_max` must also be given. The minimum of the bounding box of the mesh
        :param bbox_max: If given, `bbox_min` must also be given. The maximum of the bounding box of the mesh
        """

        super().__init__("DeformedGeneratedTextureCoordinates", node_tree)

        self.group.outputs.new("NodeSocketVector", "Vector")

        self.geometry_node = self.add_node("ShaderNodeNewGeometry")
        self.vec_transform_node = self.add_node("ShaderNodeVectorTransform")
        self.map_range_node = self.add_node("ShaderNodeMapRange")

        self.vec_transform_node.vector_type = "POINT"
        self.vec_transform_node.convert_from = "WORLD"
        self.vec_transform_node.convert_to = "OBJECT"

        self.map_range_node.clamp = False  # avoid clamping to 0-1
        self.map_range_node.data_type = "FLOAT_VECTOR"
        self.register_bounds(mesh, bbox_min, bbox_max)

        # link everything up
        self.link(
            self.geometry_node.outputs["Position"],
            self.vec_transform_node.inputs["Vector"],
        )
        self.link(
            self.vec_transform_node.outputs["Vector"],
            self.map_range_node.inputs["Vector"],
        )
        self.link(
            self.map_range_node.outputs["Vector"], self.output_node.inputs["Vector"]
        )

        self.tidy()

    def register_bounds(
        self, mesh: types.Mesh, bbox_min: types.VectorLike, bbox_max: types.VectorLike
    ):
        # calculate bbox_min, bbox_max if not given
        if mesh:
            bbox_min, bbox_max = mesh.get_raw_bounds()

        self.map_range_node.inputs[7].default_value = bbox_min
        self.map_range_node.inputs[8].default_value = bbox_max



import bpy
from bpy_extras.object_utils import world_to_camera_view
import mathutils
import numpy as np


def project_point_to_image(
    P: mathutils.Vector,
    scene: bpy.types.Scene,
    camera: bpy.types.Camera,
    invert_y: bool = True,
):
    """Return 2D (x, y) image coordinates of a 3D point P.

    :param P: 3D point
    :param scene: Blender scene
    :param camera: Camera to project through
    :param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)
    """

    ndc_x, ndc_y, ndc_depth = world_to_camera_view(
        scene, camera, P
    )  # normalized device coords
    imgx = scene.render.resolution_x * ndc_x

    if invert_y:
        imgy = scene.render.resolution_y * (1 - ndc_y)  # ndc_y measured from bottom
    else:
        imgy = scene.render.resolution_y * ndc_y

    return imgx, imgy


def project_points(
    points: np.ndarray,
    scene: bpy.types.Scene,
    camera: bpy.types.Camera,
    invert_y: bool = True,
):
    """Project points to image plane.

    :param points: Nx3 matrix of points
    :param scene: Blender scene
    :param camera: Camera to project through
    :param invert_y: if True, y is measured from the top of the image, otherwise from the bottom

    :return: Nx2 matrix of 2D image coordinates"""
    coords_2d = []
    for p in points:
        p_vec = mathutils.Vector(*(p,))
        coords_2d.append(
            project_point_to_image(p_vec, scene, camera, invert_y=invert_y)
        )

    return np.array(coords_2d)



"""Blender device management."""

import bpy
import warnings
from typing import List


# no-inherited-members


def _isgpu(device):
    return device.type in {"CUDA", "OPENCL", "METAL"}


class DeviceList(list):
    @property
    def names(self) -> List[str]:
        """Return the names of all devices."""
        return [device.name for device in self]


class Devices:
    """A manager for Blender rendering devices (CPU & GPU)."""

    def __init__(self):
        self.device_list = DeviceList(
            bpy.context.preferences.addons["cycles"].preferences.devices
        )

    def set_device_usage(
        self,
        cpu: bool = True,
        cuda: bool = True,
        opencl: bool = True,
        metal: bool = True,
    ):
        """Set/unset all devices of certain types to enabled.

        :param cpu: Enable/disable CPU devices
        :param cuda: Enable/disable CUDA devices
        :param opencl: Enable/disable OpenCL devices
        :param metal: Enable/disable Metal devices
        """

        for device in self.device_list:
            if device.type == "CPU":
                device.use = cpu

            elif device.type == "CUDA":
                device.use = cuda

            elif device.type == "OPENCL":
                device.use = opencl

            elif device.type == "METAL":
                device.use = metal

            else:
                warnings.warn(f"Unknown device type {device.type}")

    @property
    def available_gpus(self) -> DeviceList:
        """Return list of all GPU devices."""
        return DeviceList(device for device in self.device_list if _isgpu(device))

    @property
    def enabled_gpus(self) -> DeviceList:
        """Return list of available GPU devices."""
        return DeviceList(device for device in self.available_gpus if device.use)

    @property
    def device_names(self) -> List[str]:
        """Return a list of all device names."""
        return self.device_list.names

    @property
    def enabled_device_names(self) -> List[str]:
        """Return a list of named devices that are enabled."""
        return [device.name for device in self.device_list if device.use]

    def set_by_name(self, name: str, use: bool = True):
        """Set use/not use based on a device name.

        :param name: Device name
        :param use: Whether to use the device or not
        """

        for device in self.device_list:
            if device.name == name:
                device.use = use
                return

        raise KeyError(
            f"No device named {name}. Available devices: {self.device_names}"
        )



import copy

import bpy
from .utils import GetNewObject, animatable_property
from .bsyn_object import BsynObject
from ..utils import types

LIGHT_TYPES = ["POINT", "SUN", "SPOT", "AREA"]


class Light(BsynObject):
    """Light object, for managing lights in the scene"""

    light_types = LIGHT_TYPES
    """List of available light types"""

    def __init__(self, light):
        self._object = light

    @classmethod
    def from_scene(cls, key):
        """Create object from scene key

        :param key: Key of object in scene
        :return: :class:`~blendersynth.blender.light.Light` object
        """
        obj = bpy.data.objects[key]
        return cls(obj)

    @classmethod
    def create(
        cls,
        light_type: str = "POINT",
        name: str = "Light",
        intensity: float = 1.0,
        color: types.VectorLike = (1.0, 1.0, 1.0),
        location: types.VectorLike = (0, 0, 0),
    ):
        """Create a new Light object

        :param light_type: Type of light to create (see :attr:`~blendersynth.blender.light.Light.light_types`)
        :param name: Name of light
        :param intensity: Intensity of light
        :param color: Color of light
        :param location: Location of light

        :return: :class:`~blendersynth.blender.light.Light` object

        """
        assert (
            light_type in LIGHT_TYPES
        ), f"Light type `{light_type}` not found. Options are: {LIGHT_TYPES}"

        light_data = bpy.data.lights.new(name=name, type=light_type)
        light_data.energy = intensity
        light_data.color = color

        light = bpy.data.objects.new(name=name, object_data=light_data)
        light.location = location
        bpy.context.collection.objects.link(light)

        return Light(light)

    @property
    def energy(self):
        """Energy of light source"""
        return self.obj.data.energy

    @energy.setter
    def energy(self, value):
        self.set_energy(value)

    @animatable_property("energy", "data")
    def set_energy(self, value: float):
        """Set energy of light source

        :param value: Energy of light source
        :return:
        """
        self.obj.data.energy = value

    @property
    def color(self):
        """Color of light source"""
        return self.obj.data.color

    @color.setter
    def color(self, value):
        self.set_color(value)

    @animatable_property("color", "data")
    def set_color(self, value: types.VectorLike):
        """Set color of light source

        :param value: RGB[A] color
        """
        self.obj.data.color = value



"""When constructing a dataset, the INPUTS object below will
return an iterable with read's in sys.argv's `--jobs` jsons."""

import sys
import json
import os
import logging
from typing import Union, List


class Inputs:
    """This class is used to iterate over the JSONs passed in via `--jobs` in sys.argv.

    Will also convert all kwargs to attributes for easy access"""

    jsons = None
    """List of JSON files passed in via `--jobs` in sys.argv."""

    def __init__(self):
        self.jsons = sys.argv[sys.argv.index("--jobs") + 1].split(",")
        log_loc = sys.argv[sys.argv.index("--log") + 1]

        # Set up logging
        logging.basicConfig(
            filename=log_loc,
            level=logging.INFO,
            filemode="a",
            format="%(asctime)s - %(levelname)s - %(message)s",
        )

        # store all --kwargs as attributes
        for k, v in zip(sys.argv, sys.argv[1:]):
            if k.startswith("--"):
                setattr(self, k[2:], v)

    def __iter__(self):
        for n, j in enumerate(self.jsons):
            with open(j, "r") as f:
                fname = os.path.splitext(os.path.split(j)[-1])[0]
                yield fname, json.load(f)

            # Once we get here, we've passed 'yield', so we know that JSON has been loaded & rendering has occured
            logging.info(f"RENDERED: {fname}")

    def __len__(self):
        return len(self.jsons)


class DebugInputs(Inputs):
    """Class to emulate the Inputs class, for a single JSON for testing."""

    def __init__(self, json_loc: Union[str, List], repeats: int = 1):
        """

        :param json_loc: Location of JSON(s) to use
        :param repeat: Number of times to repeat the JSON(s)
        """
        if isinstance(json_loc, str):
            json_loc = [json_loc]

        self.jsons = json_loc * repeats



import bpy

AREA_TYPES = [
    "VIEW_3D",
    "IMAGE_EDITOR",
    "NODE_EDITOR",
    "SEQUENCE_EDITOR",
    "CLIP_EDITOR",
    "DOPESHEET_EDITOR",
    "GRAPH_EDITOR",
    "NLA_EDITOR",
    "TEXT_EDITOR",
    "CONSOLE",
    "INFO",
    "TOPBAR",
    "STATUSBAR",
    "OUTLINER",
    "PROPERTIES",
    "FILE_BROWSER",
    "SPREADSHEET",
    "PREFERENCES",
]
"""List of all area types"""


def get_areas(area_type):
    """Yield all areas of the given type.

    :param area_type: Area type to get (one of :attr:`~blendersynth.utils.layout.AREA_TYPES`)
    """
    assert area_type in AREA_TYPES, f"Invalid area type: {area_type}"

    for area in bpy.context.screen.areas:
        if area.type == area_type:
            yield area


def get_area(area_type) -> bpy.types.Area:
    """Return the first found area of the given type.

    :param area_type: Area type to get (one of :attr:`~blendersynth.utils.layout.AREA_TYPES`)
    """
    for area in get_areas(area_type):
        return area


def change_area_to(area_type_from: str, area_type_to: str):
    """Change the first area found.

    :param area_type_from: Area type to change from (one of :attr:`~blendersynth.utils.layout.AREA_TYPES`)
    :param area_type_to: Area type to change to (one of :attr:`~blendersynth.utils.layout.AREA_TYPES`)
    """
    get_area(area_type_from).type = area_type_to



from setuptools import setup, find_packages, Extension

setup(
	name='blendersynth',
	version='0.2.2',
	install_requires=["platformdirs>=3.11.0", "numpy", "tqdm", "ffmpeg-python"],
	packages=find_packages(include=['blendersynth', 'blendersynth.*']),
)


import bpy


def load_blend(src):
    """Load a .blend file into the current blender session.

    Note: This can cause context issues, so use with caution. If possible, use `blendersynth.run_this_script` instead.
    """

    def fix_context():
        """Fix bpy.context if some command (like .blend import) changed/emptied it"""
        for window in bpy.context.window_manager.windows:
            screen = window.screen
            for area in screen.areas:
                if area.type == "VIEW_3D":
                    for region in area.regions:
                        if region.type == "WINDOW":
                            override = {
                                "window": window,
                                "screen": screen,
                                "area": area,
                                "region": region,
                            }
                            bpy.ops.screen.screen_full_area(override)
                            break

    f = bpy.ops.wm.open_mainfile(filepath=src)
    fix_context()
    return f



"""Import python module - if not installed, install it"""

from .blender_setup.check_blender_install import _install_module
import sys


def import_module(named_module, pip_name=None):
    """Import python module - if not installed, install it.

    :param named_module: name of module to import (python name)
    :param pip_name: name of module to install (pip name). If None, will use named_module
    """

    if pip_name is None:
        pip_name = named_module

    try:
        return __import__(named_module)
    except ModuleNotFoundError:
        print(
            f"Blender executable does not have module `{named_module}` installed. Installing..."
        )
        python_exec = sys.executable
        _install_module(python_exec, pip_name)
        return __import__(named_module)



"""Overlay RGB image in compositor"""
from ..nodes import CompositorNodeGroup
import bpy
import cv2
import numpy as np
from ...file.tempfiles import create_temp_file
from typing import List
from ...utils import io


class AlphaImageOverlay(CompositorNodeGroup):
    """Overlay an image on top of the render, using the alpha channel of the image as a mask"""

    def __init__(self, name="AlphaImageOverlay", node_tree=None, scene=None):
        """Create a mix node which overlays an image on top of the input image."""

        super().__init__(name, node_tree)

        # Set default width and height
        self.width = 1000
        self.height = 1000
        if scene:
            self.width = scene.render.resolution_x
            self.height = scene.render.resolution_y

        # define I/O
        self.add_socket("NodeSocketColor", "Image", "INPUT")
        self.add_socket("NodeSocketColor", "Image", "OUTPUT")

        # create nodes
        self.overlay_img = self.group.nodes.new("CompositorNodeImage")
        self.mix_node = self.group.nodes.new("CompositorNodeMixRGB")
        self.sep_color_node = self.group.nodes.new("CompositorNodeSepRGBA")

        # link up internal nodes
        self.group.links.new(
            self.overlay_img.outputs["Image"], self.sep_color_node.inputs["Image"]
        )

        self.group.links.new(self.input_node.outputs["Image"], self.mix_node.inputs[1])
        self.group.links.new(self.overlay_img.outputs["Image"], self.mix_node.inputs[2])
        self.group.links.new(
            self.sep_color_node.outputs[3], self.mix_node.inputs["Fac"]
        )  # alpha

        self.group.links.new(
            self.mix_node.outputs["Image"], self.output_node.inputs["Image"]
        )

        self.create_img()

        self.tidy()

    def create_img(self):
        # create temp image to draw keypoints on, with 8 random alphanumeric characters
        # 8 random alphanum chars
        self.temp_img_loc = create_temp_file(".png")

        # initialize as white
        self.img = np.ones((self.height, self.width, 3), dtype=np.uint8) * 255
        cv2.imwrite(self.temp_img_loc, self.img)

        # connect this image to the overlay node
        self.overlay_img.image = io.load_image(self.temp_img_loc)
        return self.temp_img_loc


class KeypointsOverlay(AlphaImageOverlay):
    """Overlay which draws keypoints on top of the render."""

    def __init__(
        self,
        name="KeypointsOverlay",
        node_tree=None,
        scene=None,
        camera=None,
        marker: str = "x",
        size: int = 5,
        color: tuple = (0, 0, 255),
        thickness: int = 2,
    ):
        """

        :param name:
        :param node_tree:
        :param scene:
        :param camera:
        :param marker: Marker type, either [c/circle], [s/square], [t/triangle] or [x]. Default 'x'
        :param size: Size of marker. Default 5
        :param color: Color of marker, RGB or RGBA, default (0, 0, 255) (red)
        :param thickness: Thickness of marker. Default 2
        """
        super().__init__(name, node_tree)

        self.marker = marker
        self.size = size
        self.color = color
        self.thickness = thickness

    def update(
        self,
        keypoints: np.ndarray,
        scene: bpy.types.Scene = None,
        camera: bpy.types.Camera = None,
    ):
        """Given [N x 3] keypoints, draw them onto a new temp image.

        :param keypoints: N x 3 keypoints
        :param scene: Scene to draw from
        :param camera: Camera

        """

        self.width = scene.render.resolution_x
        self.height = scene.render.resolution_y

        # reset image to black, with alpha = 0
        self.img = np.zeros((self.height, self.width, 4), dtype=np.uint8)

        color = self.color
        if len(color) == 3:
            color = (*color, 255)  # add alpha

        # draw keypoints on image
        size = self.size
        for kp in keypoints:
            if self.marker in ["c", "circle"]:
                cv2.circle(
                    self.img, (int(kp[0]), int(kp[1])), size, color, self.thickness
                )
            elif self.marker in ["s", "square"]:
                cv2.rectangle(
                    self.img,
                    (int(kp[0]) - size, int(kp[1]) - size),
                    (int(kp[0]) + size, int(kp[1]) + size),
                    color,
                    self.thickness,
                )
            elif self.marker in ["t", "triangle"]:
                points = np.array(
                    [
                        [kp[0], kp[1] - size],
                        [kp[0] - size, kp[1] + size],
                        [kp[0] + size, kp[1] + size],
                    ],
                    dtype=np.int32,
                )
                cv2.polylines(
                    self.img,
                    [points],
                    isClosed=True,
                    color=color,
                    thickness=self.thickness,
                )
                cv2.fillPoly(self.img, [points], color=color)
            elif self.marker == "x":
                cv2.line(
                    self.img,
                    (int(kp[0]) - size, int(kp[1]) - size),
                    (int(kp[0]) + size, int(kp[1]) + size),
                    color,
                    self.thickness,
                )
                cv2.line(
                    self.img,
                    (int(kp[0]) + size, int(kp[1]) - size),
                    (int(kp[0]) - size, int(kp[1]) + size),
                    color,
                    self.thickness,
                )
            else:
                raise ValueError("Invalid marker: {}".format(self.marker))

        self.save_image(self.temp_img_loc, self.img)


class BoundingBoxOverlay(AlphaImageOverlay):
    """Overlay which draws bounding boxes on top of the render."""

    def __init__(
        self,
        name="BoundingBoxOverlay",
        node_tree=None,
        scene=None,
        camera=None,
        col: tuple = (0, 0, 255, 255),
        thickness: int = 2,
    ):
        super().__init__(name, node_tree)
        self.col = col
        self.thickness = int(thickness)  # cv2 requires int

    def update(self, bboxes, scene=None, camera=None):
        """Given [N x 4] bounding boxes, draw them onto a new temp image."""

        self.width = scene.render.resolution_x
        self.height = scene.render.resolution_y

        # reset image to black, with alpha = 0
        self.img = np.zeros((self.height, self.width, 4), dtype=np.uint8)

        col = self.col
        if len(col) == 3:
            col = (col[0], col[1], col[2], 255)

        # draw bounding boxes on image
        for bbox in bboxes:
            cv2.rectangle(
                self.img,
                (int(bbox[0]), int(bbox[1])),
                (int(bbox[2]), int(bbox[3])),
                col,
                self.thickness,
            )

        self.save_image(self.temp_img_loc, self.img)


class AxesOverlay(AlphaImageOverlay):
    """Overlay which draws axes on top of an existing render"""

    def __init__(
        self,
        name="BoundingBoxOverlay",
        node_tree=None,
        scene=None,
        camera=None,
        size: int = 1,
        thickness: int = 2,
    ):
        super().__init__(name, node_tree)

        self.size = size
        self.thickness = thickness

    def update(
        self,
        points: List[np.ndarray],
        scene: bpy.types.Scene = None,
        camera: bpy.types.Camera = None,
    ):
        """Plot axes onto the image.

        :param points: list of N x 4 2D points of centre + XYZ unit axes
        :param scene: Scene
        :param camera: Camera"""

        self.width = scene.render.resolution_x
        self.height = scene.render.resolution_y
        self.img = np.zeros((self.height, self.width, 4), dtype=np.uint8)

        for P in points:
            centre = P[0].astype(int)
            for i, point in enumerate(P[1:]):
                offset = point - centre
                point2 = centre + offset * self.size
                x, y = map(int, centre)
                x2, y2 = map(int, point2)
                # note BGR color order
                # and -y because image y-axis is flipped
                self.img = cv2.line(
                    self.img,
                    (x, y),
                    (x2, y2),
                    (255 * (i == 2), 255 * (i == 1), 255 * (i == 0), 255),
                    self.thickness,
                )

        self.save_image(self.temp_img_loc, self.img)



import bpy
from .devices import Devices

ENGINES = ["BLENDER_EEVEE", "BLENDER_WORKBENCH", "CYCLES"]
"""List of valid render engines"""


def render(animation: bool = False):
    """Render the scene

    :param animation: If True, render the animation. Otherwise, render a single frame.
    """
    if animation:
        bpy.ops.render.render(animation=True)
    else:
        bpy.ops.render.render(write_still=True)


def render_with_cpu():
    """Render with CPU"""
    bpy.context.scene.cycles.device = "CPU"


def render_with_gpu(force_enable=True, silent=False):
    """Render with GPU.

    :param force_enable: If True, enable all GPU devices. Otherwise, use the devices that are already enabled. To enable separately, either change your settings in blender, or see :attr:`blendersynth.blender.devices.Devices.set_device_usage`
    :param silent: If True, do not print information about device being used.
    """

    # Needed to update the device list
    bpy.context.preferences.addons["cycles"].preferences.get_devices()

    bpy.context.scene.cycles.device = "GPU"

    devices = Devices()

    if force_enable:
        devices.set_device_usage(cpu=True, cuda=True, opencl=True, metal=True)

    enabled_gpus = devices.enabled_gpus
    if enabled_gpus:
        if not silent:
            print(f"Using GPU devices {enabled_gpus.names}")

    elif not silent:
        print(
            "No GPU devices available and enabled.\n"
            + f"Available GPUs: {devices.available_gpus.names}\n"
            + "Either set force_enable=True or enable a GPU device in Blender.\n"
            "Using CPU instead..."
        )


def set_engine(engine: str):
    """Set the render engine

    :param engine: The render engine to use. See :data:`~blendersynth.blender.render.ENGINES` for valid options.
    """
    assert engine in ENGINES, "Invalid render engine"
    bpy.context.scene.render.engine = engine


def set_resolution(x: int, y: int):
    """Set the render resolution

    :param x: The width of the image in pixels
    :param y: The height of the image in pixels"""
    bpy.context.scene.render.resolution_x = x
    bpy.context.scene.render.resolution_y = y


def set_cycles_samples(samples: int):
    """Set the number of samples for the Cycles renderer"""
    assert (
        bpy.context.scene.render.engine == "CYCLES"
    ), "Cycles must be the active render engine"
    bpy.context.scene.cycles.samples = samples


def render_depth():
    """Enable the depth pass of the renderer"""
    bpy.context.view_layer.use_pass_z = True  # enable depth pass


def set_transparent(scene=None):
    """Set the background of the scene to transparent

    :param scene: The scene to set the background of. If None, use the active scene.
    """
    if scene is None:
        scene = bpy.context.scene
    scene.render.film_transparent = True



from .bsyn_object import BsynObject
import bpy
from ..utils import types
from .utils import GetNewObject


class Empty(BsynObject):
    def __init__(self, obj: bpy.types.Object = None, name=None, **kwargs):
        if obj is None:
            obj = self._create_empty_in_blender(**kwargs)
        self._object = obj

        if name is not None:
            obj.name = name

    @classmethod
    def create(
        cls,
        location: types.VectorLike = None,
        rotation: types.VectorLike = None,
        scale: types.VectorLike = None,
        radius: float = 1.0,
        name=None,
        **kwargs
    ):
        """Create and return a new Empty instance."""
        obj = cls._create_empty_in_blender(location, rotation, scale, radius, **kwargs)
        return cls(obj, name=name)

    @staticmethod
    def _create_empty_in_blender(
        location: types.VectorLike = None,
        rotation: types.VectorLike = None,
        scale: types.VectorLike = None,
        radius: float = 1.0,
        **kwargs
    ) -> bpy.types.Object:
        """Private method to create and return a new Empty object in Blender."""

        importer = GetNewObject(bpy.context.scene)
        kwargs = {
            **kwargs,
            **{
                k: v
                for k, v in zip(
                    ["location", "rotation", "scale", "radius"],
                    [location, rotation, scale, radius],
                )
            },
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        with importer:
            bpy.ops.object.empty_add(**kwargs)

        return importer.imported_obj



import platformdirs
import os
import configparser
import shutil
import subprocess
import warnings

appname = "blendersynth"
appauthor = "BlenderSynth"
config_dir = platformdirs.user_config_dir(appname, appauthor)
os.makedirs(config_dir, exist_ok=True)
config_file = os.path.join(config_dir, "config.ini")


def is_blender_in_path() -> bool:
    """Check if blender is in the PATH (i.e. can be run from the command line with 'blender')"""
    return shutil.which("blender") is not None


def find_blender_python(blender_path: str) -> str:
    """Given a blender executable, find the python interpreter it uses

    :param blender_path: path to blender executable"""

    if read_from_config("BLENDER_PYTHON_PATH") is not None:
        return read_from_config("BLENDER_PYTHON_PATH")

    # get blender to run script to find python path
    try:
        targ_script = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "blender_python_path.py")
        )
        output = subprocess.check_output(
            [blender_path, "--background", "--python", targ_script]
        )
    except subprocess.CalledProcessError as e:
        raise Exception(f"Issues with finding blender python path. Error: {e}")

    # read output to find python path
    out = None
    for l in output.decode("utf-8").split("\n"):
        if "PYTHON INTERPRETER" in l:
            out = l.split(": ")[1].strip()
            break

    if out is not None:
        write_to_config("BLENDER_PYTHON_PATH", out)
        return out

    raise Exception("Could not find Python interpreter for Blender.")


def validate_blender_path(blender_path: str) -> bool:
    """Check if `blender_path` is a valid blender executable.

    :param blender_path: path to blender executable"""
    if os.access(blender_path, os.X_OK):
        return True

    return False


def set_blender_path(_blender_path: str = None) -> str:
    """First time set-up of Blender Path. Will try to find Blender path in following order of
    precedence:

    1) Input argument _blender_path
    2) Environment variable `BLENDER_PATH`
    3) Blender in PATH
    4) Ask user for path

    When it finds a valid path, it will save it to the Config file stored in user's `platformdirs.user_config_dir`
    """

    if _blender_path is not None:
        if validate_blender_path(_blender_path):
            blender_path = _blender_path

        elif validate_blender_path(
            _blender_path + ".exe"
        ):  # add .exe if not there (helps os.access checking)
            blender_path = _blender_path + ".exe"

        else:
            raise ValueError(
                f"Provided Blender path, {_blender_path}, is not executable."
            )

    elif os.environ.get("BLENDER_PATH") is not None:
        blender_path = os.environ.get("BLENDER_PATH")

    elif is_blender_in_path():
        blender_path = shutil.which("blender")

    else:
        blender_path = input(
            "Blender not found in PATH or Environment Variable.\nPlease provide path to blender executable: "
        )

    blender_path = os.path.abspath(blender_path)  # make sure it's absolute path

    if not validate_blender_path(blender_path):
        if validate_blender_path(blender_path + ".exe"):
            blender_path += ".exe"

        else:
            raise ValueError(
                f"Provided Blender path, {blender_path}, is not valid as a Blender executable."
            )

    write_to_config("BLENDER_PATH", blender_path)
    return blender_path


def get_blender_path() -> str:
    """Return blender path,
    or None if no config file found"""
    return read_from_config("BLENDER_PATH")


def write_to_config(key, value, section="BLENDER_SETUP"):
    """Load config, and write key value pair to cfg[section]"""
    config = configparser.ConfigParser()

    if os.path.exists(config_file):
        config.read(config_file)

    if section not in config:
        config[section] = {}

    config[section][key] = value

    with open(config_file, "w") as configfile:
        config.write(configfile)


def read_from_config(key, section="BLENDER_SETUP"):
    """Load config, and read value from cfg[section]"""
    config = configparser.ConfigParser()

    if os.path.exists(config_file):
        config.read(config_file)

    if section not in config:
        return None

    if key not in config[section]:
        return None

    return config[section][key]


def remove_from_config(key, section="BLENDER_SETUP"):
    """Load config, and remove key from cfg[section]"""
    config = configparser.ConfigParser()

    if os.path.exists(config_file):
        config.read(config_file)

    if section not in config:
        return None

    if key not in config[section]:
        return None

    del config[section][key]

    with open(config_file, "w") as configfile:
        config.write(configfile)


def remove_config():
    """Remove config file"""
    if os.path.exists(config_file):
        os.remove(config_file)
